# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pickle
import numpy as np
import os.path as op
from abc import abstractmethod
from functools import partial
from typing import List
import time, os, math, re
from packaging import version
import pkg_resources

try:
    import rdkit
    from rdkit import Chem, DataStructs
    from rdkit.Chem import AllChem, Descriptors
    import rdkit.Chem.QED as QED
    from rdkit import rdBase

    rdBase.DisableLog("rdApp.error")
    from rdkit.Chem import rdMolDescriptors
    from rdkit.six import iteritems
except:
    raise ImportError(
        "Please install rdkit by 'conda install -c conda-forge rdkit'! ")

try:
    from scipy.stats.mstats import gmean
except:
    raise ImportError("Please install rdkit by 'pip install scipy'! ")

try:
    import networkx as nx
except:
    raise ImportError("Please install networkx by 'pip install networkx'! ")

from ...utils import oracle_load
from ...utils import print_sys, install

mean2func = {
    "geometric": gmean,
    "arithmetic": np.mean,
}
SKLEARN_VERSION = version.parse(
    pkg_resources.get_distribution("scikit-learn").version)


def smiles_to_rdkit_mol(smiles):
    """Convert smiles into rdkit's mol (molecule) format.

    Args:
      smiles: str, SMILES string.

    Returns:
      mol: rdkit.Chem.rdchem.Mol

    """
    mol = Chem.MolFromSmiles(smiles)
    #  Sanitization check (detects invalid valence)
    if mol is not None:
        try:
            Chem.SanitizeMol(mol)
        except ValueError:
            return None
    return mol


def smiles_2_fingerprint_ECFP4(smiles):
    """Convert smiles into ECFP4 Morgan Fingerprint.

    Args:
      smiles: str, SMILES string.

    Returns:
      fp: rdkit.DataStructs.cDataStructs.UIntSparseIntVect

    """
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprint(molecule, 2)
    return fp


def smiles_2_fingerprint_FCFP4(smiles):
    """Convert smiles into FCFP4 Morgan Fingerprint.

    Args:
      smiles: str, SMILES string.

    Returns:
      fp: rdkit.DataStructs.cDataStructs.UIntSparseIntVect

    """
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprint(molecule, 2, useFeatures=True)
    return fp


def smiles_2_fingerprint_AP(smiles):
    """Convert smiles into Atom Pair Fingerprint.

    Args:
      smiles: str, SMILES string.

    Returns:
      fp: rdkit.DataStructs.cDataStructs.IntSparseIntVect

    """
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetAtomPairFingerprint(molecule, maxLength=10)
    return fp


def smiles_2_fingerprint_ECFP6(smiles):
    """Convert smiles into ECFP6 Fingerprint.

    Args:
      smiles: str, SMILES string.

    Returns:
      fp: rdkit.DataStructs.cDataStructs.UIntSparseIntVect

    """
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprint(molecule, 3)
    return fp


fp2fpfunc = {
    "ECFP4": smiles_2_fingerprint_ECFP4,
    "FCFP4": smiles_2_fingerprint_FCFP4,
    "AP": smiles_2_fingerprint_AP,
    "ECFP6": smiles_2_fingerprint_ECFP6,
}


class ScoreModifier:
    """
    Interface for score modifiers.
    """

    @abstractmethod
    def __call__(self, x):
        """
        Apply the modifier on x.

        Args:
            x: float or np.array to modify

        Returns:
            float or np.array (depending on the type of x) after application of the distance function.
        """


class ChainedModifier(ScoreModifier):
    """
    Calls several modifiers one after the other, for instance:
        score = modifier3(modifier2(modifier1(raw_score)))
    """

    def __init__(self, modifiers: List[ScoreModifier]) -> None:
        """
        Args:
            modifiers: modifiers to call in sequence.
                The modifier applied last (and delivering the final score) is the last one in the list.
        """
        self.modifiers = modifiers

    def __call__(self, x):
        score = x
        for modifier in self.modifiers:
            score = modifier(score)
        return score


class LinearModifier(ScoreModifier):
    """
    Score modifier that multiplies the score by a scalar (default: 1, i.e. do nothing).
    """

    def __init__(self, slope=1.0):
        self.slope = slope

    def __call__(self, x):
        return self.slope * x


class SquaredModifier(ScoreModifier):
    """
    Score modifier that has a maximum at a given target value, and decreases
    quadratically with increasing distance from the target value.
    """

    def __init__(self, target_value: float, coefficient=1.0) -> None:
        self.target_value = target_value
        self.coefficient = coefficient

    def __call__(self, x):
        return 1.0 - self.coefficient * np.square(self.target_value - x)


class AbsoluteScoreModifier(ScoreModifier):
    """
    Score modifier that has a maximum at a given target value, and decreases
    linearly with increasing distance from the target value.
    """

    def __init__(self, target_value: float) -> None:
        self.target_value = target_value

    def __call__(self, x):
        return 1.0 - np.abs(self.target_value - x)


class GaussianModifier(ScoreModifier):
    """
    Score modifier that reproduces a Gaussian bell shape.
    """

    def __init__(self, mu: float, sigma: float) -> None:
        self.mu = mu
        self.sigma = sigma

    def __call__(self, x):
        return np.exp(-0.5 * np.power((x - self.mu) / self.sigma, 2.0))


class MinMaxGaussianModifier(ScoreModifier):
    """
    Score modifier that reproduces a half Gaussian bell shape.
    For minimize==True, the function is 1.0 for x <= mu and decreases to zero for x > mu.
    For minimize==False, the function is 1.0 for x >= mu and decreases to zero for x < mu.
    """

    def __init__(self, mu: float, sigma: float, minimize=False) -> None:
        self.mu = mu
        self.sigma = sigma
        self.minimize = minimize
        self._full_gaussian = GaussianModifier(mu=mu, sigma=sigma)

    def __call__(self, x):
        if self.minimize:
            mod_x = np.maximum(x, self.mu)
        else:
            mod_x = np.minimum(x, self.mu)
        return self._full_gaussian(mod_x)


MinGaussianModifier = partial(MinMaxGaussianModifier, minimize=True)
MaxGaussianModifier = partial(MinMaxGaussianModifier, minimize=False)


class ClippedScoreModifier(ScoreModifier):
    r"""
    Clips a score between specified low and high scores, and does a linear interpolation in between.

    This class works as follows:
    First the input is mapped onto a linear interpolation between both specified points.
    Then the generated values are clipped between low and high scores.
    """

    def __init__(self,
                 upper_x: float,
                 lower_x=0.0,
                 high_score=1.0,
                 low_score=0.0) -> None:
        """
        Args:
            upper_x: x-value from which (or until which if smaller than lower_x) the score is maximal
            lower_x: x-value until which (or from which if larger than upper_x) the score is minimal
            high_score: maximal score to clip to
            low_score: minimal score to clip to
        """
        assert low_score < high_score

        self.upper_x = upper_x
        self.lower_x = lower_x
        self.high_score = high_score
        self.low_score = low_score

        self.slope = (high_score - low_score) / (upper_x - lower_x)
        self.intercept = high_score - self.slope * upper_x

    def __call__(self, x):
        y = self.slope * x + self.intercept
        return np.clip(y, self.low_score, self.high_score)


class SmoothClippedScoreModifier(ScoreModifier):
    """
    Smooth variant of ClippedScoreModifier.

    Implemented as a logistic function that has the same steepness as ClippedScoreModifier in the
    center of the logistic function.
    """

    def __init__(self,
                 upper_x: float,
                 lower_x=0.0,
                 high_score=1.0,
                 low_score=0.0) -> None:
        """
        Args:
            upper_x: x-value from which (or until which if smaller than lower_x) the score approaches high_score
            lower_x: x-value until which (or from which if larger than upper_x) the score approaches low_score
            high_score: maximal score (reached at +/- infinity)
            low_score: minimal score (reached at -/+ infinity)
        """
        assert low_score < high_score

        self.upper_x = upper_x
        self.lower_x = lower_x
        self.high_score = high_score
        self.low_score = low_score

        # Slope of a standard logistic function in the middle is 0.25 -> rescale k accordingly
        self.k = 4.0 / (upper_x - lower_x)
        self.middle_x = (upper_x + lower_x) / 2
        self.L = high_score - low_score

    def __call__(self, x):
        return self.low_score + self.L / (1 + np.exp(-self.k *
                                                     (x - self.middle_x)))


class ThresholdedLinearModifier(ScoreModifier):
    """
    Returns a value of min(input, threshold)/threshold.
    """

    def __init__(self, threshold: float) -> None:
        self.threshold = threshold

    def __call__(self, x):
        return np.minimum(x, self.threshold) / self.threshold


# check the license for the code from readFragmentScores to CalculateScore here: https://github.com/EricTing/SAscore/blob/89d7689a85efed3cc918fb8ba6fe5cedf60b4a5a/src/sascorer.py#L134
_fscores = None


def readFragmentScores(name="fpscores"):
    import gzip

    global _fscores
    # generate the full path filename:
    # if name == "fpscores":
    #     name = op.join(previous_directory(op.dirname(__file__)), name)
    name = oracle_load("fpscores")
    try:
        with open("oracle/fpscores.pkl", "rb") as f:
            _fscores = pickle.load(f)
    except EOFError:
        import sys

        sys.exit(
            "TDC is hosted in Harvard Dataverse and it is currently under maintenance, please check back in a few hours or checkout https://dataverse.harvard.edu/."
        )

    outDict = {}
    for i in _fscores:
        for j in range(1, len(i)):
            outDict[i[j]] = float(i[0])
    _fscores = outDict


def numBridgeheadsAndSpiro(mol, ri=None):
    nSpiro = rdMolDescriptors.CalcNumSpiroAtoms(mol)
    nBridgehead = rdMolDescriptors.CalcNumBridgeheadAtoms(mol)
    return nBridgehead, nSpiro


def calculateScore(m):
    if _fscores is None:
        readFragmentScores()

    # fragment score
    fp = rdMolDescriptors.GetMorganFingerprint(
        m, 2)  # <- 2 is the *radius* of the circular fingerprint
    fps = fp.GetNonzeroElements()
    score1 = 0.0
    nf = 0
    for bitId, v in iteritems(fps):
        nf += v
        sfp = bitId
        score1 += _fscores.get(sfp, -4) * v
    score1 /= nf

    # features score
    nAtoms = m.GetNumAtoms()
    nChiralCenters = len(Chem.FindMolChiralCenters(m, includeUnassigned=True))
    ri = m.GetRingInfo()
    nBridgeheads, nSpiro = numBridgeheadsAndSpiro(m, ri)
    nMacrocycles = 0
    for x in ri.AtomRings():
        if len(x) > 8:
            nMacrocycles += 1

    sizePenalty = nAtoms**1.005 - nAtoms
    stereoPenalty = math.log10(nChiralCenters + 1)
    spiroPenalty = math.log10(nSpiro + 1)
    bridgePenalty = math.log10(nBridgeheads + 1)
    macrocyclePenalty = 0.0
    # ---------------------------------------
    # This differs from the paper, which defines:
    #  macrocyclePenalty = math.log10(nMacrocycles+1)
    # This form generates better results when 2 or more macrocycles are present
    if nMacrocycles > 0:
        macrocyclePenalty = math.log10(2)

    score2 = (0.0 - sizePenalty - stereoPenalty - spiroPenalty - bridgePenalty -
              macrocyclePenalty)

    # correction for the fingerprint density
    # not in the original publication, added in version 1.1
    # to make highly symmetrical molecules easier to synthetise
    score3 = 0.0
    if nAtoms > len(fps):
        score3 = math.log(float(nAtoms) / len(fps)) * 0.5

    sascore = score1 + score2 + score3

    # need to transform "raw" value into scale between 1 and 10
    min = -4.0
    max = 2.5
    sascore = 11.0 - (sascore - min + 1) / (max - min) * 9.0
    # smooth the 10-end
    if sascore > 8.0:
        sascore = 8.0 + math.log(sascore + 1.0 - 9.0)
    if sascore > 10.0:
        sascore = 10.0
    elif sascore < 1.0:
        sascore = 1.0

    return sascore


"""Scores based on an ECFP classifier for activity."""


def load_pickled_model(name: str):
    """
    Loading a pretrained model serialized with pickle.
    Usually for sklearn models.

    Args:
      name: Name of the model to load.

    Returns:
      The model.
    """

    try:
        with open(name, "rb") as f:
            model = pickle.load(f)
    except EOFError:
        import sys

        sys.exit(
            "TDC is hosted in Harvard Dataverse and it is currently under maintenance, please check back in a few hours or checkout https://dataverse.harvard.edu/."
        )
    return model


# clf_model = None
def load_drd2_model():
    name = "oracle/drd2.pkl"

    if SKLEARN_VERSION >= version.parse("0.24.0"):
        name = "oracle/drd2_current.pkl"
    else:
        name = "oracle/drd2.pkl"

    return load_pickled_model(name)


def fingerprints_from_mol(mol):
    fp = AllChem.GetMorganFingerprint(mol, 3, useCounts=True, useFeatures=True)
    size = 2048
    nfp = np.zeros((1, size), np.int32)
    for idx, v in fp.GetNonzeroElements().items():
        nidx = idx % size
        nfp[0, nidx] += int(v)
    return nfp


def drd2(smile):
    """Evaluate DRD2 score of a SMILES string

    Args:
      smiles: str

    Returns:
      drd_score: float

    """

    if "drd2_model" not in globals().keys():
        global drd2_model
        drd2_model = load_drd2_model()

    mol = Chem.MolFromSmiles(smile)
    if mol:
        fp = fingerprints_from_mol(mol)
        score = drd2_model.predict_proba(fp)[:, 1]
        drd_score = float(score)
        return drd_score
    return 0.0


def load_cyp3a4_veith():
    oracle_file = "oracle/cyp3a4_veith.pkl"
    return load_pickled_model(oracle_file)


def cyp3a4_veith(smiles):
    try:
        from DeepPurpose import utils
    except:
        raise ImportError(
            "Please install DeepPurpose by 'pip install DeepPurpose'")

    import os

    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    if "cyp3a4_veith_model" not in globals().keys():
        global cyp3a4_veith_model
        cyp3a4_veith_model = load_cyp3a4_veith()

    import warnings, os

    warnings.filterwarnings("ignore")

    X_drug = [smiles]
    drug_encoding = "CNN"
    y = [1]
    X_pred = utils.data_process(X_drug=X_drug,
                                y=y,
                                drug_encoding=drug_encoding,
                                split_method="no_split")
    # cyp3a4_veith_model = cyp3a4_veith_model.to("cuda:0")
    y_pred = cyp3a4_veith_model.predict(X_pred)
    return y_pred[0]


## from https://github.com/wengong-jin/iclr19-graph2graph/blob/master/props/properties.py
## from https://github.com/wengong-jin/multiobj-rationale/blob/master/properties.py


def similarity(smiles_a, smiles_b):
    """Evaluate Tanimoto similarity between 2 SMILES strings

    Args:
      smiles_a: str, SMILES string
      smiles_b: str, SMILES string

    Returns:
      similarity score: float, between 0 and 1.

    """
    if smiles_a is None or smiles_b is None:
        return 0.0
    amol = Chem.MolFromSmiles(smiles_a)
    bmol = Chem.MolFromSmiles(smiles_b)
    if amol is None or bmol is None:
        return 0.0
    fp1 = AllChem.GetMorganFingerprintAsBitVect(amol,
                                                2,
                                                nBits=2048,
                                                useChirality=False)
    fp2 = AllChem.GetMorganFingerprintAsBitVect(bmol,
                                                2,
                                                nBits=2048,
                                                useChirality=False)
    return DataStructs.TanimotoSimilarity(fp1, fp2)


def qed(smiles):
    """Evaluate QED score of a SMILES string

    Args:
      smiles: str

    Returns:
      qed_score: float, between 0 and 1.

    """
    if smiles is None:
        return 0.0
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return 0.0
    return QED.qed(mol)


def penalized_logp(s):
    """Evaluate LogP score of a SMILES string

    Args:
      smiles: str

    Returns:
      logp_score: float, between - infinity and + infinity

    """
    if s is None:
        return -100.0
    mol = Chem.MolFromSmiles(s)
    if mol is None:
        return -100.0

    logP_mean = 2.4570953396190123
    logP_std = 1.434324401111988
    SA_mean = -3.0525811293166134
    SA_std = 0.8335207024513095
    cycle_mean = -0.0485696876403053
    cycle_std = 0.2860212110245455
    log_p = Descriptors.MolLogP(mol)
    # SA = -sascorer.calculateScore(mol)
    SA = -calculateScore(mol)

    # cycle score
    cycle_list = nx.cycle_basis(nx.Graph(Chem.rdmolops.GetAdjacencyMatrix(mol)))
    if len(cycle_list) == 0:
        cycle_length = 0
    else:
        cycle_length = max([len(j) for j in cycle_list])
    if cycle_length <= 6:
        cycle_length = 0
    else:
        cycle_length = cycle_length - 6
    cycle_score = -cycle_length

    normalized_log_p = (log_p - logP_mean) / logP_std
    normalized_SA = (SA - SA_mean) / SA_std
    normalized_cycle = (cycle_score - cycle_mean) / cycle_std
    return normalized_log_p + normalized_SA + normalized_cycle


def SA(s):
    """Evaluate SA score of a SMILES string

    Args:
      smiles: str

    Returns:
      SAscore: float

    """
    if s is None:
        return 100
    mol = Chem.MolFromSmiles(s)
    if mol is None:
        return 100
    SAscore = calculateScore(mol)
    return SAscore


def load_gsk3b_model():
    gsk3_model_path = "oracle/gsk3b.pkl"
    if SKLEARN_VERSION >= version.parse("0.24.0"):
        gsk3_model_path = "oracle/gsk3b_current.pkl"
    return load_pickled_model(gsk3_model_path)


def gsk3b(smiles):
    """Evaluate GSK3B score of a SMILES string

    Args:
      smiles: str

    Returns:
      gsk3_score: float, between 0 and 1.

    """
    if "gsk3_model" not in globals().keys():
        global gsk3_model
        gsk3_model = load_gsk3b_model()

    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 2, nBits=2048)
    features = np.zeros((1,))
    DataStructs.ConvertToNumpyArray(fp, features)
    fp = features.reshape(1, -1)
    gsk3_score = gsk3_model.predict_proba(fp)[0, 1]
    return gsk3_score


class jnk3:
    """Evaluate JSK3 score of a SMILES string

    Args:
      smiles: str

    Returns:
      jnk3_score: float , between 0 and 1.

    """

    def __init__(self):

        jnk3_model_path = "oracle/jnk3.pkl"
        if SKLEARN_VERSION >= version.parse("0.24.0"):
            jnk3_model_path = "oracle/jnk3_current.pkl"
        self.jnk3_model = load_pickled_model(jnk3_model_path)

    def __call__(self, smiles):
        molecule = smiles_to_rdkit_mol(smiles)
        fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 2, nBits=2048)
        features = np.zeros((1,))
        DataStructs.ConvertToNumpyArray(fp, features)
        fp = features.reshape(1, -1)
        jnk3_score = self.jnk3_model.predict_proba(fp)[0, 1]
        return jnk3_score


class AtomCounter:

    def __init__(self, element):
        """
        Args:
            element: element to count within a molecule
        """
        self.element = element

    def __call__(self, mol):
        """
        Count the number of atoms of a given type.

        Args:
            mol: molecule

        Returns:
            The number of atoms of the given type.
        """
        # if the molecule contains H atoms, they may be implicit, so add them
        if self.element == "H":
            mol = Chem.AddHs(mol)

        return sum(1 for a in mol.GetAtoms() if a.GetSymbol() == self.element)


def parse_molecular_formula(formula):
    """
    Parse a molecular formulat to get the element types and counts.

    Args:
        formula: molecular formula, f.i. "C8H3F3Br"

    Returns:
        A list of tuples containing element types and number of occurrences.
    """
    import re

    matches = re.findall(r"([A-Z][a-z]*)(\d*)", formula)

    # Convert matches to the required format
    results = []
    for match in matches:
        # convert count to an integer, and set it to 1 if the count is not visible in the molecular formula
        count = 1 if not match[1] else int(match[1])
        results.append((match[0], count))

    return results


def smiles2formula(smiles):

    from rdkit.Chem.rdMolDescriptors import CalcMolFormula

    mol = Chem.MolFromSmiles(smiles)
    formula = CalcMolFormula(mol)
    return formula


def canonicalize(smiles: str, include_stereocenters=True):
    """
    Canonicalize the SMILES strings with RDKit.

    The algorithm is detailed under https://pubs.acs.org/doi/full/10.1021/acs.jcim.5b00543

    Args:
        smiles: SMILES string to canonicalize
        include_stereocenters: whether to keep the stereochemical information in the canonical SMILES string

    Returns:
        Canonicalized SMILES string, None if the molecule is invalid.
    """

    mol = Chem.MolFromSmiles(smiles)

    if mol is not None:
        return Chem.MolToSmiles(mol, isomericSmiles=include_stereocenters)
    else:
        return None


class Isomer_scoring_prev:

    def __init__(self, target_smiles, means="geometric"):
        assert means in ["geometric", "arithmetic"]
        if means == "geometric":
            self.mean_func = gmean
        else:
            self.mean_func = np.mean
        atom2cnt_lst = parse_molecular_formula(target_smiles)
        total_atom_num = sum([cnt for atom, cnt in atom2cnt_lst])
        self.total_atom_modifier = GaussianModifier(mu=total_atom_num,
                                                    sigma=2.0)
        self.AtomCounter_Modifier_lst = [((AtomCounter(atom)),
                                          GaussianModifier(mu=cnt, sigma=1.0))
                                         for atom, cnt in atom2cnt_lst]

    def __call__(self, test_smiles):
        molecule = smiles_to_rdkit_mol(test_smiles)
        all_scores = []
        for atom_counter, modifier_func in self.AtomCounter_Modifier_lst:
            all_scores.append(modifier_func(atom_counter(molecule)))

        ### total atom number
        atom2cnt_lst = parse_molecular_formula(test_smiles)
        # ## todo add Hs
        total_atom_num = sum([cnt for atom, cnt in atom2cnt_lst])
        all_scores.append(self.total_atom_modifier(total_atom_num))
        return self.mean_func(all_scores)


class Isomer_scoring:

    def __init__(self, target_smiles, means="geometric"):
        assert means in ["geometric", "arithmetic"]
        if means == "geometric":
            self.mean_func = gmean
        else:
            self.mean_func = np.mean
        atom2cnt_lst = parse_molecular_formula(target_smiles)
        total_atom_num = sum([cnt for atom, cnt in atom2cnt_lst])
        self.total_atom_modifier = GaussianModifier(mu=total_atom_num,
                                                    sigma=2.0)
        self.AtomCounter_Modifier_lst = [((AtomCounter(atom)),
                                          GaussianModifier(mu=cnt, sigma=1.0))
                                         for atom, cnt in atom2cnt_lst]

    def __call__(self, test_smiles):
        #### difference 1
        #### add hydrogen atoms
        test_smiles = canonicalize(test_smiles)
        test_mol = Chem.MolFromSmiles(test_smiles)
        test_mol2 = Chem.AddHs(test_mol)
        test_smiles = Chem.MolToSmiles(test_mol2)

        molecule = smiles_to_rdkit_mol(test_smiles)
        all_scores = []
        for atom_counter, modifier_func in self.AtomCounter_Modifier_lst:
            all_scores.append(modifier_func(atom_counter(molecule)))

        #### difference 2
        ### total atom number
        test_formula = smiles2formula(test_smiles)
        atom2cnt_lst = parse_molecular_formula(test_formula)
        # atom2cnt_lst = parse_molecular_formula(test_smiles)
        # ## todo add Hs
        total_atom_num = sum([cnt for atom, cnt in atom2cnt_lst])
        all_scores.append(self.total_atom_modifier(total_atom_num))
        return self.mean_func(all_scores)


def isomer_meta_prev(target_smiles, means="geometric"):
    return Isomer_scoring_prev(target_smiles, means=means)


def isomer_meta(target_smiles, means="geometric"):
    return Isomer_scoring(target_smiles, means=means)


isomers_c7h8n2o2_prev = isomer_meta_prev(target_smiles="C7H8N2O2",
                                         means="geometric")
isomers_c9h10n2o2pf2cl_prev = isomer_meta_prev(target_smiles="C9H10N2O2PF2Cl",
                                               means="geometric")
isomers_c11h24_prev = isomer_meta_prev(target_smiles="C11H24",
                                       means="geometric")

isomers_c7h8n2o2 = isomer_meta(target_smiles="C7H8N2O2", means="geometric")
isomers_c9h10n2o2pf2cl = isomer_meta(target_smiles="C9H10N2O2PF2Cl",
                                     means="geometric")
isomers_c11h24 = isomer_meta(target_smiles="C11H24", means="geometric")


class rediscovery_meta:

    def __init__(self, target_smiles, fp="ECFP4"):
        self.similarity_func = fp2fpfunc[fp]
        self.target_fp = self.similarity_func(target_smiles)

    def __call__(self, test_smiles):
        test_fp = self.similarity_func(test_smiles)
        similarity_value = DataStructs.TanimotoSimilarity(
            self.target_fp, test_fp)
        return similarity_value


class similarity_meta:

    def __init__(self, target_smiles, fp="FCFP4", modifier_func=None):
        self.similarity_func = fp2fpfunc[fp]
        self.target_fp = self.similarity_func(target_smiles)
        self.modifier_func = modifier_func

    def __call__(self, test_smiles):
        test_fp = self.similarity_func(test_smiles)
        similarity_value = DataStructs.TanimotoSimilarity(
            self.target_fp, test_fp)
        if self.modifier_func is None:
            modifier_score = similarity_value
        else:
            modifier_score = self.modifier_func(similarity_value)
        return modifier_score


celecoxib_rediscovery = rediscovery_meta(
    target_smiles="CC1=CC=C(C=C1)C1=CC(=NN1C1=CC=C(C=C1)S(N)(=O)=O)C(F)(F)F",
    fp="ECFP4")
troglitazone_rediscovery = rediscovery_meta(
    target_smiles="Cc1c(C)c2OC(C)(COc3ccc(CC4SC(=O)NC4=O)cc3)CCc2c(C)c1O",
    fp="ECFP4")
thiothixene_rediscovery = rediscovery_meta(
    target_smiles="CN(C)S(=O)(=O)c1ccc2Sc3ccccc3C(=CCCN4CCN(C)CC4)c2c1",
    fp="ECFP4")

similarity_modifier = ClippedScoreModifier(upper_x=0.75)
aripiprazole_similarity = similarity_meta(
    target_smiles="Clc4cccc(N3CCN(CCCCOc2ccc1c(NC(=O)CC1)c2)CC3)c4Cl",
    fp="FCFP4",
    modifier_func=similarity_modifier,
)

albuterol_similarity = similarity_meta(
    target_smiles="CC(C)(C)NCC(O)c1ccc(O)c(CO)c1",
    fp="FCFP4",
    modifier_func=similarity_modifier,
)

mestranol_similarity = similarity_meta(
    target_smiles="COc1ccc2[C@H]3CC[C@@]4(C)[C@@H](CC[C@@]4(O)C#C)[C@@H]3CCc2c1",
    fp="AP",
    modifier_func=similarity_modifier,
)


class median_meta:

    def __init__(
        self,
        target_smiles_1,
        target_smiles_2,
        fp1="ECFP6",
        fp2="ECFP6",
        modifier_func1=None,
        modifier_func2=None,
        means="geometric",
    ):
        self.similarity_func1 = fp2fpfunc[fp1]
        self.similarity_func2 = fp2fpfunc[fp2]
        self.target_fp1 = self.similarity_func1(target_smiles_1)
        self.target_fp2 = self.similarity_func2(target_smiles_2)
        self.modifier_func1 = modifier_func1
        self.modifier_func2 = modifier_func2
        assert means in ["geometric", "arithmetic"]
        self.mean_func = mean2func[means]

    def __call__(self, test_smiles):
        test_fp1 = self.similarity_func1(test_smiles)
        test_fp2 = (test_fp1 if self.similarity_func2 == self.similarity_func1
                    else self.similarity_func2(test_smiles))
        similarity_value1 = DataStructs.TanimotoSimilarity(
            self.target_fp1, test_fp1)
        similarity_value2 = DataStructs.TanimotoSimilarity(
            self.target_fp2, test_fp2)
        if self.modifier_func1 is None:
            modifier_score1 = similarity_value1
        else:
            modifier_score1 = self.modifier_func1(similarity_value1)
        if self.modifier_func2 is None:
            modifier_score2 = similarity_value2
        else:
            modifier_score2 = self.modifier_func2(similarity_value2)
        final_score = self.mean_func([modifier_score1, modifier_score2])
        return final_score


camphor_smiles = "CC1(C)C2CCC1(C)C(=O)C2"
menthol_smiles = "CC(C)C1CCC(C)CC1O"

median1 = median_meta(
    target_smiles_1=camphor_smiles,
    target_smiles_2=menthol_smiles,
    fp1="ECFP4",
    fp2="ECFP4",
    modifier_func1=None,
    modifier_func2=None,
    means="geometric",
)

tadalafil_smiles = "O=C1N(CC(N2C1CC3=C(C2C4=CC5=C(OCO5)C=C4)NC6=C3C=CC=C6)=O)C"
sildenafil_smiles = "CCCC1=NN(C2=C1N=C(NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"
median2 = median_meta(
    target_smiles_1=tadalafil_smiles,
    target_smiles_2=sildenafil_smiles,
    fp1="ECFP6",
    fp2="ECFP6",
    modifier_func1=None,
    modifier_func2=None,
    means="geometric",
)


class MPO_meta:

    def __init__(self, means):
        """
        target_smiles, fp in ['ECFP4', 'AP', ..., ]
        scoring,
        modifier,

        """

        assert means in ["geometric", "arithmetic"]
        self.mean_func = mean2func[means]

    def __call__(self, test_smiles):
        molecule = smiles_to_rdkit_mol(test_smiles)

        score_lst = []
        return self.mean_func(score_lst)


def osimertinib_mpo(test_smiles):

    if "osimertinib_fp_fcfc4" not in globals().keys():
        global osimertinib_fp_fcfc4, osimertinib_fp_ecfc6
        osimertinib_smiles = (
            "COc1cc(N(C)CCN(C)C)c(NC(=O)C=C)cc1Nc2nccc(n2)c3cn(C)c4ccccc34")
        osimertinib_fp_fcfc4 = smiles_2_fingerprint_FCFP4(osimertinib_smiles)
        osimertinib_fp_ecfc6 = smiles_2_fingerprint_ECFP6(osimertinib_smiles)

    sim_v1_modifier = ClippedScoreModifier(upper_x=0.8)
    sim_v2_modifier = MinGaussianModifier(mu=0.85, sigma=0.1)
    tpsa_modifier = MaxGaussianModifier(mu=100, sigma=10)
    logp_modifier = MinGaussianModifier(mu=1, sigma=1)

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp_fcfc4 = smiles_2_fingerprint_FCFP4(test_smiles)
    fp_ecfc6 = smiles_2_fingerprint_ECFP6(test_smiles)
    tpsa_score = tpsa_modifier(Descriptors.TPSA(molecule))
    logp_score = logp_modifier(Descriptors.MolLogP(molecule))
    similarity_v1 = sim_v1_modifier(
        DataStructs.TanimotoSimilarity(osimertinib_fp_fcfc4, fp_fcfc4))
    similarity_v2 = sim_v2_modifier(
        DataStructs.TanimotoSimilarity(osimertinib_fp_ecfc6, fp_ecfc6))

    osimertinib_gmean = gmean(
        [tpsa_score, logp_score, similarity_v1, similarity_v2])
    return osimertinib_gmean


def fexofenadine_mpo(test_smiles):
    if "fexofenadine_fp" not in globals().keys():
        global fexofenadine_fp
        fexofenadine_smiles = (
            "CC(C)(C(=O)O)c1ccc(cc1)C(O)CCCN2CCC(CC2)C(O)(c3ccccc3)c4ccccc4")
        fexofenadine_fp = smiles_2_fingerprint_AP(fexofenadine_smiles)

    similar_modifier = ClippedScoreModifier(upper_x=0.8)
    tpsa_modifier = MaxGaussianModifier(mu=90, sigma=10)
    logp_modifier = MinGaussianModifier(mu=4, sigma=1)

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp_ap = smiles_2_fingerprint_AP(test_smiles)
    tpsa_score = tpsa_modifier(Descriptors.TPSA(molecule))
    logp_score = logp_modifier(Descriptors.MolLogP(molecule))
    similarity_value = similar_modifier(
        DataStructs.TanimotoSimilarity(fp_ap, fexofenadine_fp))
    fexofenadine_gmean = gmean([tpsa_score, logp_score, similarity_value])
    return fexofenadine_gmean


def ranolazine_mpo(test_smiles):
    if "ranolazine_fp" not in globals().keys():
        global ranolazine_fp, fluorine_counter
        ranolazine_smiles = "COc1ccccc1OCC(O)CN2CCN(CC(=O)Nc3c(C)cccc3C)CC2"
        ranolazine_fp = smiles_2_fingerprint_AP(ranolazine_smiles)
        fluorine_counter = AtomCounter("F")

    similar_modifier = ClippedScoreModifier(upper_x=0.7)
    tpsa_modifier = MaxGaussianModifier(mu=95, sigma=20)
    logp_modifier = MaxGaussianModifier(mu=7, sigma=1)
    fluorine_modifier = GaussianModifier(mu=1, sigma=1.0)

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp_ap = smiles_2_fingerprint_AP(test_smiles)
    tpsa_score = tpsa_modifier(Descriptors.TPSA(molecule))
    logp_score = logp_modifier(Descriptors.MolLogP(molecule))
    similarity_value = similar_modifier(
        DataStructs.TanimotoSimilarity(fp_ap, ranolazine_fp))
    fluorine_value = fluorine_modifier(fluorine_counter(molecule))

    ranolazine_gmean = gmean(
        [tpsa_score, logp_score, similarity_value, fluorine_value])
    return ranolazine_gmean


def perindopril_mpo(test_smiles):
    ## no similar_modifier

    if "perindopril_fp" not in globals().keys():
        global perindopril_fp, num_aromatic_rings
        perindopril_smiles = "O=C(OCC)C(NC(C(=O)N1C(C(=O)O)CC2CCCCC12)C)CCC"
        perindopril_fp = smiles_2_fingerprint_ECFP4(perindopril_smiles)

        def num_aromatic_rings(mol):
            return rdMolDescriptors.CalcNumAromaticRings(mol)

    arom_rings_modifier = GaussianModifier(mu=2, sigma=0.5)

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp_ecfp4 = smiles_2_fingerprint_ECFP4(test_smiles)

    similarity_value = DataStructs.TanimotoSimilarity(fp_ecfp4, perindopril_fp)
    num_aromatic_rings_value = arom_rings_modifier(num_aromatic_rings(molecule))

    perindopril_gmean = gmean([similarity_value, num_aromatic_rings_value])
    return perindopril_gmean


def amlodipine_mpo(test_smiles):
    ## no similar_modifier
    if "amlodipine_fp" not in globals().keys():
        global amlodipine_fp, num_rings
        amlodipine_smiles = "Clc1ccccc1C2C(=C(/N/C(=C2/C(=O)OCC)COCCN)C)\C(=O)OC"
        amlodipine_fp = smiles_2_fingerprint_ECFP4(amlodipine_smiles)

        def num_rings(mol):
            return rdMolDescriptors.CalcNumRings(mol)

    num_rings_modifier = GaussianModifier(mu=3, sigma=0.5)

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp_ecfp4 = smiles_2_fingerprint_ECFP4(test_smiles)

    similarity_value = DataStructs.TanimotoSimilarity(fp_ecfp4, amlodipine_fp)
    num_rings_value = num_rings_modifier(num_rings(molecule))

    amlodipine_gmean = gmean([similarity_value, num_rings_value])
    return amlodipine_gmean


def zaleplon_mpo_prev(test_smiles):
    if "zaleplon_fp" not in globals().keys():
        global zaleplon_fp, isomer_scoring_C19H17N3O2
        zaleplon_smiles = "O=C(C)N(CC)C1=CC=CC(C2=CC=NC3=C(C=NN23)C#N)=C1"
        zaleplon_fp = smiles_2_fingerprint_ECFP4(zaleplon_smiles)
        isomer_scoring_C19H17N3O2 = Isomer_scoring_prev(
            target_smiles="C19H17N3O2")

    fp = smiles_2_fingerprint_ECFP4(test_smiles)
    similarity_value = DataStructs.TanimotoSimilarity(fp, zaleplon_fp)
    isomer_value = isomer_scoring_C19H17N3O2(test_smiles)
    return gmean([similarity_value, isomer_value])


def zaleplon_mpo(test_smiles):
    if "zaleplon_fp" not in globals().keys():
        global zaleplon_fp, isomer_scoring_C19H17N3O2
        zaleplon_smiles = "O=C(C)N(CC)C1=CC=CC(C2=CC=NC3=C(C=NN23)C#N)=C1"
        zaleplon_fp = smiles_2_fingerprint_ECFP4(zaleplon_smiles)
        isomer_scoring_C19H17N3O2 = Isomer_scoring(target_smiles="C19H17N3O2")

    fp = smiles_2_fingerprint_ECFP4(test_smiles)
    similarity_value = DataStructs.TanimotoSimilarity(fp, zaleplon_fp)
    isomer_value = isomer_scoring_C19H17N3O2(test_smiles)
    return gmean([similarity_value, isomer_value])


def sitagliptin_mpo_prev(test_smiles):
    if "sitagliptin_fp_ecfp4" not in globals().keys():
        global sitagliptin_fp_ecfp4, sitagliptin_logp_modifier, sitagliptin_tpsa_modifier, isomers_scoring_C16H15F6N5O, sitagliptin_similar_modifier
        sitagliptin_smiles = "Fc1cc(c(F)cc1F)CC(N)CC(=O)N3Cc2nnc(n2CC3)C(F)(F)F"
        sitagliptin_fp_ecfp4 = smiles_2_fingerprint_ECFP4(sitagliptin_smiles)
        sitagliptin_mol = Chem.MolFromSmiles(sitagliptin_smiles)
        sitagliptin_logp = Descriptors.MolLogP(sitagliptin_mol)
        sitagliptin_tpsa = Descriptors.TPSA(sitagliptin_mol)
        sitagliptin_logp_modifier = GaussianModifier(mu=sitagliptin_logp,
                                                     sigma=0.2)
        sitagliptin_tpsa_modifier = GaussianModifier(mu=sitagliptin_tpsa,
                                                     sigma=5)
        isomers_scoring_C16H15F6N5O = Isomer_scoring_prev("C16H15F6N5O")
        sitagliptin_similar_modifier = GaussianModifier(mu=0, sigma=0.1)

    molecule = Chem.MolFromSmiles(test_smiles)
    fp_ecfp4 = smiles_2_fingerprint_ECFP4(test_smiles)
    logp_score = Descriptors.MolLogP(molecule)
    logp_score = sitagliptin_logp_modifier(logp_score)
    tpsa_score = Descriptors.TPSA(molecule)
    tpsa_score = sitagliptin_tpsa_modifier(tpsa_score)
    isomer_score = isomers_scoring_C16H15F6N5O(test_smiles)
    similarity_value = sitagliptin_similar_modifier(
        DataStructs.TanimotoSimilarity(fp_ecfp4, sitagliptin_fp_ecfp4))
    return gmean([similarity_value, logp_score, tpsa_score, isomer_score])


def sitagliptin_mpo(test_smiles):
    if "sitagliptin_fp_ecfp4" not in globals().keys():
        global sitagliptin_fp_ecfp4, sitagliptin_logp_modifier, sitagliptin_tpsa_modifier, isomers_scoring_C16H15F6N5O, sitagliptin_similar_modifier
        sitagliptin_smiles = "Fc1cc(c(F)cc1F)CC(N)CC(=O)N3Cc2nnc(n2CC3)C(F)(F)F"
        sitagliptin_fp_ecfp4 = smiles_2_fingerprint_ECFP4(sitagliptin_smiles)
        sitagliptin_mol = Chem.MolFromSmiles(sitagliptin_smiles)
        sitagliptin_logp = Descriptors.MolLogP(sitagliptin_mol)
        sitagliptin_tpsa = Descriptors.TPSA(sitagliptin_mol)
        sitagliptin_logp_modifier = GaussianModifier(mu=sitagliptin_logp,
                                                     sigma=0.2)
        sitagliptin_tpsa_modifier = GaussianModifier(mu=sitagliptin_tpsa,
                                                     sigma=5)
        isomers_scoring_C16H15F6N5O = Isomer_scoring("C16H15F6N5O")
        sitagliptin_similar_modifier = GaussianModifier(mu=0, sigma=0.1)

    molecule = Chem.MolFromSmiles(test_smiles)
    fp_ecfp4 = smiles_2_fingerprint_ECFP4(test_smiles)
    logp_score = Descriptors.MolLogP(molecule)
    logp_score = sitagliptin_logp_modifier(logp_score)
    tpsa_score = Descriptors.TPSA(molecule)
    tpsa_score = sitagliptin_tpsa_modifier(tpsa_score)
    isomer_score = isomers_scoring_C16H15F6N5O(test_smiles)
    similarity_value = sitagliptin_similar_modifier(
        DataStructs.TanimotoSimilarity(fp_ecfp4, sitagliptin_fp_ecfp4))
    return gmean([similarity_value, logp_score, tpsa_score, isomer_score])


def get_PHCO_fingerprint(mol):
    if "Gobbi_Pharm2D" not in globals().keys():
        global Gobbi_Pharm2D, Generate
        from rdkit.Chem.Pharm2D import Generate, Gobbi_Pharm2D
    return Generate.Gen2DFingerprint(mol, Gobbi_Pharm2D.factory)


class SMARTS_scoring:

    def __init__(self, target_smarts, inverse):
        self.target_mol = Chem.MolFromSmarts(target_smarts)
        self.inverse = inverse

    def __call__(self, mol):
        matches = mol.GetSubstructMatches(self.target_mol)
        if len(matches) > 0:
            if self.inverse:
                return 0.0
            else:
                return 1.0
        else:
            if self.inverse:
                return 1.0
            else:
                return 0.0


def deco_hop(test_smiles):
    if "pharmacophor_fp" not in globals().keys():
        global pharmacophor_fp, deco1_smarts_scoring, deco2_smarts_scoring, scaffold_smarts_scoring
        pharmacophor_smiles = "CCCOc1cc2ncnc(Nc3ccc4ncsc4c3)c2cc1S(=O)(=O)C(C)(C)C"
        pharmacophor_mol = smiles_to_rdkit_mol(pharmacophor_smiles)
        pharmacophor_fp = get_PHCO_fingerprint(pharmacophor_mol)

        deco1_smarts_scoring = SMARTS_scoring(target_smarts="CS([#6])(=O)=O",
                                              inverse=True)
        deco2_smarts_scoring = SMARTS_scoring(
            target_smarts="[#7]-c1ccc2ncsc2c1", inverse=True)
        scaffold_smarts_scoring = SMARTS_scoring(
            target_smarts="[#7]-c1n[c;h1]nc2[c;h1]c(-[#8])[c;h0][c;h1]c12",
            inverse=False,
        )

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp = get_PHCO_fingerprint(molecule)
    similarity_modifier = ClippedScoreModifier(upper_x=0.85)

    similarity_value = similarity_modifier(
        DataStructs.TanimotoSimilarity(fp, pharmacophor_fp))
    deco1_score = deco1_smarts_scoring(molecule)
    deco2_score = deco2_smarts_scoring(molecule)
    scaffold_score = scaffold_smarts_scoring(molecule)

    all_scores = np.mean(
        [similarity_value, deco1_score, deco2_score, scaffold_score])
    return all_scores


def scaffold_hop(test_smiles):
    if ("pharmacophor_fp" not in globals().keys() or
            "scaffold_smarts_scoring" not in globals().keys() or
            "deco_smarts_scoring" not in globals().keys()):
        global pharmacophor_fp, deco_smarts_scoring, scaffold_smarts_scoring
        pharmacophor_smiles = "CCCOc1cc2ncnc(Nc3ccc4ncsc4c3)c2cc1S(=O)(=O)C(C)(C)C"
        pharmacophor_mol = smiles_to_rdkit_mol(pharmacophor_smiles)
        pharmacophor_fp = get_PHCO_fingerprint(pharmacophor_mol)

        deco_smarts_scoring = SMARTS_scoring(
            target_smarts=
            "[#6]-[#6]-[#6]-[#8]-[#6]~[#6]~[#6]~[#6]~[#6]-[#7]-c1ccc2ncsc2c1",
            inverse=False,
        )

        scaffold_smarts_scoring = SMARTS_scoring(
            target_smarts="[#7]-c1n[c;h1]nc2[c;h1]c(-[#8])[c;h0][c;h1]c12",
            inverse=True)

    molecule = smiles_to_rdkit_mol(test_smiles)
    fp = get_PHCO_fingerprint(molecule)
    similarity_modifier = ClippedScoreModifier(upper_x=0.75)

    similarity_value = similarity_modifier(
        DataStructs.TanimotoSimilarity(fp, pharmacophor_fp))
    deco_score = deco_smarts_scoring(molecule)
    scaffold_score = scaffold_smarts_scoring(molecule)

    all_scores = np.mean([similarity_value, deco_score, scaffold_score])
    return all_scores


def valsartan_smarts(test_smiles):
    if "valsartan_logp_modifier" not in globals().keys():
        global valsartan_mol, valsartan_logp_modifier, valsartan_tpsa_modifier, valsartan_bertz_modifier
        valsartan_smarts = "CN(C=O)Cc1ccc(c2ccccc2)cc1"  ### smarts
        valsartan_mol = Chem.MolFromSmarts(valsartan_smarts)

        sitagliptin_smiles = (
            "NC(CC(=O)N1CCn2c(nnc2C(F)(F)F)C1)Cc1cc(F)c(F)cc1F"  ### other mol
        )
        sitagliptin_mol = Chem.MolFromSmiles(sitagliptin_smiles)

        target_logp = Descriptors.MolLogP(sitagliptin_mol)
        target_tpsa = Descriptors.TPSA(sitagliptin_mol)
        target_bertz = Descriptors.BertzCT(sitagliptin_mol)

        valsartan_logp_modifier = GaussianModifier(mu=target_logp, sigma=0.2)
        valsartan_tpsa_modifier = GaussianModifier(mu=target_tpsa, sigma=5)
        valsartan_bertz_modifier = GaussianModifier(mu=target_bertz, sigma=30)

    molecule = smiles_to_rdkit_mol(test_smiles)
    matches = molecule.GetSubstructMatches(valsartan_mol)
    if len(matches) > 0:
        smarts_score = 1.0
    else:
        smarts_score = 0.0

    logp_score = valsartan_logp_modifier(Descriptors.MolLogP(molecule))
    tpsa_score = valsartan_tpsa_modifier(Descriptors.TPSA(molecule))
    bertz_score = valsartan_bertz_modifier(Descriptors.BertzCT(molecule))
    valsartan_gmean = gmean([smarts_score, tpsa_score, logp_score, bertz_score])
    return valsartan_gmean


###########################################################################
###               END of Guacamol
###########################################################################
"""
Synthesizability from a full retrosynthetic analysis
Including:
    1. MIT ASKCOS
    ASKCOS (https://askcos.mit.edu) is an open-source software 
    framework that integrates efforts to generalize known chemistry 
    to new substrates by learning to apply retrosynthetic transformations, 
    to identify suitable reaction conditions, and to evaluate whether 
    reactions are likely to be successful. The data-driven models are trained 
    with USPTO and Reaxys databases.
    
    Reference:
    https://doi.org/10.1021/acs.jcim.0c00174

    2. IBM_RXN
    IBM RXN (https://rxn.res.ibm.com) is an AI platform integarting 
    forward reaction prediction and retrosynthetic analysis. The 
    backend of the IBM RXN retrosynthetic analysis is Molecular 
    Transformer model (see reference). The model was mainly trained 
    with USPTO, Pistachio databases.
    Reference:
    https://doi.org/10.1021/acscentsci.9b00576
"""


def tree_analysis(current):
    """
    Analyze the result of tree builder
    Calculate: 1. Number of steps 2. \Pi plausibility 3. If find a path
    In case of celery error, all values are -1

    return:
        num_path = number of paths found
        status: Same as implemented in ASKCOS one
        num_step: number of steps
        p_score: \Pi plausibility
        synthesizability: binary code
        price: price for synthesize query compound
    """
    if "error" in current.keys():
        return -1, {}, 11, -1, -1, -1

    if "price" in current.keys():
        return 0, {}, 0, 1, 1, current["price"]

    num_path = len(current["trees"])
    if num_path != 0:
        current = [current["trees"][0]]
        if current[0]["ppg"] != 0:
            return 0, {}, 0, 1, 1, current[0]["ppg"]
    else:
        current = []

    depth = 0
    p_score = 1
    status = {0: 1}
    price = 0
    while True:
        num_child = 0
        depth += 0.5
        temp = []
        for i, item in enumerate(current):
            num_child += len(item["children"])
            temp = temp + item["children"]
        if num_child == 0:
            break
        if depth % 1 != 0:
            for sth in temp:
                p_score = p_score * sth["plausibility"]
        else:
            for mol in temp:
                price += mol["ppg"]
        status[depth] = num_child
        current = temp
    if len(status) > 1:
        synthesizability = 1
    else:
        synthesizability = 0
    if int(depth - 0.5) == 0:
        depth = 11
        price = -1
    else:
        depth = int(depth - 0.5)
    return num_path, status, depth, p_score * synthesizability, synthesizability, price


def askcos(
    smiles,
    host_ip,
    output="plausibility",
    save_json=False,
    file_name="tree_builder_result.json",
    num_trials=5,
    max_depth=9,
    max_branching=25,
    expansion_time=60,
    max_ppg=100,
    template_count=1000,
    max_cum_prob=0.999,
    chemical_property_logic="none",
    max_chemprop_c=0,
    max_chemprop_n=0,
    max_chemprop_o=0,
    max_chemprop_h=0,
    chemical_popularity_logic="none",
    min_chempop_reactants=5,
    min_chempop_products=5,
    filter_threshold=0.1,
    return_first="true",
):
    """
    The ASKCOS retrosynthetic analysis oracle function.
    Please refer https://github.com/connorcoley/ASKCOS to run the ASKCOS with docker on a server to receive requests.
    """

    if output not in ["num_step", "plausibility", "synthesizability", "price"]:
        raise NameError(
            "This output value is not implemented. Please select one from 'num_step', 'plausibility', 'synthesizability', 'price'."
        )

    import json, requests

    params = {"smiles": smiles}
    resp = requests.get(host_ip + "/api/price/", params=params, verify=False)

    if resp.json()["price"] == 0:
        # Parameters for Tree Builder
        params = {
            "smiles": smiles,
            # optional
            "max_depth": max_depth,
            "max_branching": max_branching,
            "expansion_time": expansion_time,
            "max_ppg": max_ppg,
            "template_count": template_count,
            "max_cum_prob": max_cum_prob,
            "chemical_property_logic": chemical_property_logic,
            "max_chemprop_c": max_chemprop_c,
            "max_chemprop_n": max_chemprop_n,
            "max_chemprop_o": max_chemprop_o,
            "max_chemprop_h": max_chemprop_h,
            "chemical_popularity_logic": chemical_popularity_logic,
            "min_chempop_reactants": min_chempop_reactants,
            "min_chempop_products": min_chempop_products,
            "filter_threshold": filter_threshold,
            "return_first": return_first,
        }

        # For each entry, repeat to test up to num_trials times if got error message
        for _ in range(num_trials):
            print("Trying to send the request, for the %i times now" % (_ + 1))
            resp = requests.get(host_ip + "/api/treebuilder/",
                                params=params,
                                verify=False)
            if "error" not in resp.json().keys():
                break

    if save_json:
        with open(file_name, "w") as f_data:
            json.dump(resp.json(), f_data)

    num_path, status, depth, p_score, synthesizability, price = tree_analysis(
        resp.json())

    if output == "plausibility":
        return p_score
    elif output == "num_step":
        return depth
    elif output == "synthesizability":
        return synthesizability
    elif output == "price":
        return price


def ibm_rxn(smiles, api_key, output="confidence", sleep_time=30):
    """
    This function is modified from Dr. Jan Jensen's code
    """
    try:
        from rxn4chemistry import RXN4ChemistryWrapper
    except:
        print_sys("Please install rxn4chemistry via pip install rxn4chemistry")
    import time

    rxn4chemistry_wrapper = RXN4ChemistryWrapper(api_key=api_key)
    response = rxn4chemistry_wrapper.create_project("test")
    time.sleep(sleep_time)
    response = rxn4chemistry_wrapper.predict_automatic_retrosynthesis(
        product=smiles)
    status = ""
    while status != "SUCCESS":
        time.sleep(sleep_time)
        results = rxn4chemistry_wrapper.get_predict_automatic_retrosynthesis_results(
            response["prediction_id"])
        status = results["status"]

    if output == "confidence":
        return results["retrosynthetic_paths"][0]["confidence"]
    elif output == "result":
        return results
    else:
        raise NameError("This output value is not implemented.")


class molecule_one_retro:

    def __init__(self, api_token):
        try:
            from m1wrapper import MoleculeOneWrapper
        except:
            try:
                install(
                    "git+https://github.com/molecule-one/m1wrapper-python@v1")
                from m1wrapper import MoleculeOneWrapper
            except:
                raise ImportError(
                    "Install Molecule.One Wrapper via pip install git+https://github.com/molecule-one/m1wrapper-python@v1"
                )

        self.m1wrapper = MoleculeOneWrapper(api_token,
                                            "https://tdc.molecule.one")

    def __call__(self, smiles):
        if isinstance(smiles, str):
            smiles = [smiles]

        search = self.m1wrapper.run_batch_search(
            targets=smiles,
            parameters={
                "exploratory_search": False,
                "detail_level": "score"
            },
        )

        status_cur = search.get_status()
        print_sys("Started Querying...")
        print_sys(status_cur)
        while True:
            time.sleep(7)
            status = search.get_status()

            if (status["queued"] == 0) and (status["running"] == 0):
                print_sys("Finished... Returning Results...")
                break
            else:
                if status_cur != status:
                    print_sys(status)
            status_cur = status
        result = search.get_results(precision=5,
                                    only=["targetSmiles", "result"])
        return {i["targetSmiles"]: i["result"] for i in result}


class PyScreener_meta:
    """Evaluate docking score

    Args:

    Return:


    """

    def __init__(
        self,
        receptor_pdb_file,
        box_center,
        box_size,
        software_class="vina",
        ncpu=4,
        **kwargs,
    ):
        try:
            import ray

            try:
                ray.init()
            except:
                ray.shutdown()
                ray.init()
            import pyscreener as ps
        except:
            raise ImportError(
                "Please install PyScreener following guidance in https://github.com/coleygroup/pyscreener"
            )

        try:
            metadata = ps.build_metadata(software_class)
        except:
            raise ValueError(
                'The value of software_class is not implemented. Currently available:["vina", "qvina", "smina", "psovina", "dock", "dock6", "ucsfdock"]'
            )

        self.scorer = ps.virtual_screen(
            software_class,
            [receptor_pdb_file],
            box_center,
            box_size,
            metadata,
            ncpu=ncpu,
        )

    def __call__(self, test_smiles, error_value=None):
        final_score = self.scorer(test_smiles)
        if type(test_smiles) == str:
            return list(final_score)[0]
        else:  ## list
            # dict: {'O=C(/C=C/c1ccc([N+](=O)[O-])o1)c1ccc(-c2ccccc2)cc1': -9.9, 'CCOc1cc(/C=C/C(=O)C(=Cc2ccc(O)c(OC)c2)C(=O)/C=C/c2ccc(O)c(OCC)c2)ccc1O': -9.1}
            # return [list(i.values())[0] for i in final_score]
            score_lst = []
            for smiles in test_smiles:
                score = final_score[smiles]
                if score is None:
                    score = error_value
                score_lst.append(score)
            return score_lst


class Score_3d:
    """Evaluate Vina score (force field) for a conformer binding to a receptor"""

    def __init__(self,
                 receptor_pdbqt_file,
                 center,
                 box_size,
                 scorefunction="vina"):
        try:
            from vina import Vina
        except:
            raise ImportError(
                "Please install vina following guidance in https://github.com/ccsb-scripps/AutoDock-Vina/tree/develop/build/python"
            )

        self.v = Vina(sf_name=scorefunction)
        self.receptor_pdbqt_file = receptor_pdbqt_file
        self.center = center
        self.box_size = box_size
        self.v.set_receptor(rigid_pdbqt_filename=receptor_pdbqt_file)
        try:
            self.v.compute_vina_maps(center=self.center, box_size=self.box_size)
        except:
            raise ValueError(
                "Cannot compute the affinity map, please check center and box_size"
            )

    def __call__(self, ligand_pdbqt_file, minimize=True):
        try:
            self.v.set_ligand_from_file(ligand_pdbqt_file)
            if minimize:
                energy = self.v.optimize()[0]
            else:
                energy = self.v.score()[0]
        except Exception as e:
            print(e)
            return np.inf
        return energy


class Vina_3d:
    """Perform docking search from a conformer."""

    def __init__(self,
                 receptor_pdbqt_file,
                 center,
                 box_size,
                 scorefunction="vina"):
        try:
            from vina import Vina
        except:
            raise ImportError(
                "Please install vina following guidance in https://github.com/ccsb-scripps/AutoDock-Vina/tree/develop/build/python"
            )

        self.v = Vina(sf_name=scorefunction)
        self.receptor_pdbqt_file = receptor_pdbqt_file
        self.center = center
        self.box_size = box_size
        self.v.set_receptor(rigid_pdbqt_filename=receptor_pdbqt_file)
        try:
            self.v.compute_vina_maps(center=self.center, box_size=self.box_size)
        except:
            raise ValueError(
                "Cannot compute the affinity map, please check center and box_size"
            )

    def __call__(self,
                 ligand_pdbqt_file,
                 output_file="out.pdbqt",
                 exhaustiveness=8,
                 n_poses=10):
        try:
            self.v.set_ligand_from_file(ligand_pdbqt_file)
            self.v.dock(exhaustiveness=exhaustiveness, n_poses=n_poses)
            self.v.write_poses(output_file, n_poses=n_poses, overwrite=True)
            energy = self.v.score()[0]
        except Exception as e:
            print(e)
            return np.inf
        return energy


class Vina_smiles:
    """Perform docking search from a conformer."""

    def __init__(self,
                 receptor_pdbqt_file,
                 center,
                 box_size,
                 scorefunction="vina"):
        try:
            from vina import Vina
        except:
            raise ImportError(
                "Please install vina following guidance in https://github.com/ccsb-scripps/AutoDock-Vina/tree/develop/build/python"
            )

        self.v = Vina(sf_name=scorefunction)
        self.receptor_pdbqt_file = receptor_pdbqt_file
        self.center = center
        self.box_size = box_size
        self.v.set_receptor(rigid_pdbqt_filename=receptor_pdbqt_file)
        # self.v.set_receptor(rigid_pdbqt_filename = "tdc/receptors/1iep/1iep_receptor.pdbqt")
        try:
            self.v.compute_vina_maps(center=self.center, box_size=self.box_size)
        except:
            raise ValueError(
                "Cannot compute the affinity map, please check center and box_size"
            )

    def __call__(self,
                 ligand_smiles,
                 output_file="out.pdbqt",
                 exhaustiveness=8,
                 n_poses=10):
        try:
            m = Chem.MolFromSmiles(ligand_smiles)
            m = Chem.AddHs(m)
            AllChem.EmbedMolecule(m)
            AllChem.MMFFOptimizeMolecule(m)
            print(Chem.MolToMolBlock(m), file=open("__temp.mol", "w+"))
            os.system("mk_prepare_ligand.py -i __temp.mol -o __temp.pdbqt")
            self.v.set_ligand_from_file("__temp.pdbqt")
            self.v.dock(exhaustiveness=exhaustiveness, n_poses=n_poses)
            self.v.write_poses(output_file, n_poses=n_poses, overwrite=True)
            energy = self.v.score()[0]
            os.system("rm __temp.mol __temp.pdbqt")
        except Exception as e:
            print(e)
            return np.inf
        return energy


def smina(ligand, protein, score_only=False, raw_input=False):
    """
    Sima is a docking algorithm that docks a ligand to a protein pocket.

    Koes, D.R., Baumgartner, M.P. and Camacho, C.J., 2013.
    Lessons learned in empirical scoring with smina from the CSAR 2011 benchmarking exercise.
    Journal of chemical information and modeling, 53(8), pp.1893-1904.

    Parameters
    ----------
    ligand : array
        (N_1,3) matrix, where N_1 is ligand size.
    protein : array
        (N_2,3) matrix, where N_2 is protein size.
    score_only: boolean
        whether to only return docking score.
    raw_input: boolean
        whether to input raw ML input or sdf file input
    Returns
    -------
    docking_info: str or float
        docking result

    """
    smina_model_path = "oracle/smina.static"
    os.system(f"chmod +x ./{smina_model_path}")
    # if machine learning raw input:
    # 1. write out to xyz file
    # 2. convert to sdf file by openbabel
    if raw_input:
        mol_coord, mol_atom = ligand
        # 1. write out to xyz file
        f = open(f"temp_ligand.xyz", "w")
        n_atoms = len(mol_atom)
        f.write("%d\n\n" % n_atoms)
        for atom_i in range(n_atoms):
            atom = mol_atom[atom_i]
            f.write("%s %.9f %.9f %.9f\n" % (
                atom,
                mol_coord[atom_i, 0],
                mol_coord[atom_i, 1],
                mol_coord[atom_i, 2],
            ))
        f.close()
        # 2. convert to sdf file
        try:
            os.system(f"obabel temp_ligand.xyz -O temp_ligand.sdf")
        except:
            raise ImportError(
                "Please install openbabel by 'conda install -c conda-forge openbabel'!"
            )
        ligand = "temp_ligand.sdf"
    if score_only:
        msg = os.popen(
            f"./{smina_model_path} -l {ligand} -r {protein} --score_only").read(
            )
        return float(msg.split("\n")[-7].split(" ")[-2])
    else:
        os.system(f"./{smina_model_path} -l {ligand} -r {protein} --score_only")


# os.system("python docking.py " + ligand_pdbqt_file + \
#           " "+target_pdbqt_file + " " + output_file +' '+ \
#           docking_center_string + ' ' + box_size_string)



"""General class for representing dataset configs"""

import pandas as pd

from ..base_dataset import DataLoader
from ..feature_generators import resource, base


class ConfigBase(dict):
    """Class for representing configs"""
    pass


class DatasetConfig(ConfigBase):
    """Class for representing dataset configs"""

    def __init__(self,
                 dataset_name=None,
                 data_processing_class=None,
                 functions_to_run=None,
                 args_for_functions=None,
                 var_map=None,
                 **kwargs):
        self.dataset_name = dataset_name
        self.data_processing_class = data_processing_class
        self.functions_to_run = functions_to_run
        self.args_for_functions = args_for_functions
        self.var_map = var_map
        self["kwargs"] = kwargs

    @property
    def name(self):
        return self.dataset_name

    @property
    def processing_class(self):
        return self.data_processing_class

    @property
    def functions(self):
        return self.functions_to_run

    @property
    def args(self):
        return self.args_for_functions

    @property
    def functions_and_args(self):
        return zip(self.functions, self.args)

    @property
    def processing_callback(self):
        """Returns a callback for processing a dataset according to this config"""
        return (lambda dataset: self.data_processing_class.process_data(
            dataset, self.functions, self.args)
               ) if self.data_processing_class is not None else None

    @property
    def tdc_cols_callback(self, dataset_type=None):
        """Add the standard  TDC columns (e.g., 'X1', 'X2', ...) to a given DataFrame."""

        def bi_pred_var_map_helper(dataset):
            assert dataset is not None, "Must provide a dataset."
            dataset["X1"] = dataset[self.var_map["X1"]]
            dataset["X2"] = dataset[self.var_map["X2"]]
            dataset["ID1"] = dataset[self.var_map["ID1"]]
            dataset["ID2"] = dataset[self.var_map["ID2"]]
            return dataset

        dataset_type = dataset_type if dataset_type is not None else "bi_pred_dataset"
        if dataset_type == "bi_pred_dataset" and self.var_map is not None:
            return bi_pred_var_map_helper
        else:
            return None


class LoaderConfig(DatasetConfig):
    """Class for representing configs for DataLoader instances
    
    Requires DataLoader parent class to inherit from dictionary
    LoaderConfig allows DataLoader instance variables to be used
    as parameters in feature generation"""

    def __init__(self,
                 keys=None,
                 loader_functions=None,
                 loader_args=None,
                 feature_class=None,
                 dataset_name=None,
                 data_processing_class=None,
                 functions_to_run=None,
                 args_for_functions=None,
                 var_map=None,
                 **kwargs):
        if feature_class is not None:
            assert isinstance(feature_class, base.FeatureGenerator)
            self.feature_class = feature_class
        super().__init__(dataset_name=None,
                         data_processing_class=None,
                         functions_to_run=None,
                         args_for_functions=None,
                         var_map=None,
                         **kwargs)
        self.loader_functions = loader_functions if loader_functions else []
        self.loader_args = loader_args if loader_args else []
        self.keys = keys

    @property
    def loader_setup_callback(self):
        """Returns a callback for processing a dataloader according to this config"""
        return lambda loader: self.feature_class.process_loader(
            loader, self.keys, self.loader_functions, self.loader_args
        ) if self.feature_class is not None else None


class ResourceConfig(LoaderConfig):
    """Class for representing configs for Resource DataLoader instances"""

    def __init__(self,
                 feature_class,
                 dataset_name=None,
                 data_processing_class=None,
                 functions_to_run=None,
                 args_for_functions=None,
                 var_map=None,
                 **kwargs):
        assert isinstance(feature_class, resource.ResourceFeatureGenerator)
        super().__init__(feature_class=feature_class, **kwargs)

    @property
    def df_key(self):
        return "df"

    @property
    def split_key(self):
        return "split"



from .evaluator import (
    validity,
    uniqueness,
    novelty,
    diversity,
    kl_divergence,
    fcd_distance,
)
from .featurize.molconvert import MolConvert
from .oracle.oracle import (
    PyScreener_meta,
    Vina_3d,
    Score_3d,
    Vina_smiles,
    molecule_one_retro,
    ibm_rxn,
    askcos,
    isomers_c7h8n2o2,
    isomers_c9h10n2o2pf2cl,
    isomers_c11h24,
    valsartan_smarts,
    scaffold_hop,
    deco_hop,
    sitagliptin_mpo,
    zaleplon_mpo,
    amlodipine_mpo,
    sitagliptin_mpo_prev,
    zaleplon_mpo_prev,
    perindopril_mpo,
    ranolazine_mpo,
    fexofenadine_mpo,
    osimertinib_mpo,
    median1,
    median2,
    aripiprazole_similarity,
    albuterol_similarity,
    mestranol_similarity,
    celecoxib_rediscovery,
    troglitazone_rediscovery,
    thiothixene_rediscovery,
    median_meta,
    isomer_meta,
    rediscovery_meta,
    similarity_meta,
    jnk3,
    gsk3b,
    SA,
    cyp3a4_veith,
    drd2,
    qed,
    penalized_logp,
)
from .oracle.filter import MolFilter



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import sys
import warnings

warnings.filterwarnings("ignore")

from . import single_pred_dataset
from ..utils import print_sys, fuzzy_search, property_dataset_load
from ..metadata import dataset_names


class ADME(single_pred_dataset.DataLoader):
    """Data loader class to load datasets in ADME task. More info: https://tdcommons.ai/single_pred_tasks/adme/

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False
        convert_format (str, optional):
            Automatic conversion of SMILES to other molecular formats in MolConvert class. Stored as separate column in dataframe, defaults to None
    """

    def __init__(
        self,
        name,
        path="./data",
        label_name=None,
        print_stats=False,
        convert_format=None,
    ):
        """Create ADME dataloader object."""
        super().__init__(
            name,
            path,
            label_name,
            print_stats,
            dataset_names=dataset_names["ADME"],
            convert_format=convert_format,
        )

        self.name = fuzzy_search(self.name, dataset_names["ADME"])
        if self.name == "ppbr_az":
            import pandas as pd
            import os

            self.ppbr_df = pd.read_csv(os.path.join(self.path,
                                                    self.name + ".tab"),
                                       sep="\t")
            df = self.ppbr_df[self.ppbr_df.Species == "Homo sapiens"]
            self.entity1 = df.Drug.values
            self.y = df.Y.values
            self.entity1_idx = df.Drug_ID.values

        if print_stats:
            self.print_stats()
        print("Done!", flush=True, file=sys.stderr)

    def get_approved_set(self):
        import pandas as pd

        if self.name not in ["pampa_ncats"]:
            raise ValueError(
                "This function is only available for PAMPA_NCATS dataset")
        entity1, y, entity1_idx = property_dataset_load("approved_pampa_ncats",
                                                        self.path, None,
                                                        dataset_names["ADME"])
        return pd.DataFrame({"Drug_ID": entity1_idx, "Drug": entity1, "Y": y})

    def get_other_species(self, species=None):

        if self.name not in ["ppbr_az"]:
            raise ValueError(
                "This function is only available for assays with species label, including PPBR"
            )

        if species == "all":
            return self.ppbr_df

        if species in self.ppbr_df.Species.unique():
            return self.ppbr_df[self.ppbr_df.Species == species].reset_index(
                drop=True)
        else:
            raise ValueError(
                "You can only specify the following set of species name: 'Canis lupus familiaris', 'Cavia porcellus', 'Homo sapiens', 'Mus musculus', 'Rattus norvegicus', 'all'"
            )

    def harmonize(self, mode=None):
        """Removing duplicated experimental readouts."""

        if mode not in ["max", "min", "remove_all"]:
            raise ValueError(
                "Please specify 'mode' of removal, currently supported 'max'/'min'/'remove_all'!"
            )

        if mode == "max":
            df_ = self.get_data()
            df = (df_.sort_values(
                "Y",
                ascending=True).drop_duplicates("Drug").reset_index(drop=True))

        elif mode == "min":
            df_ = self.get_data()
            df = (df_.sort_values(
                "Y",
                ascending=False).drop_duplicates("Drug").reset_index(drop=True))

        elif mode == "remove_all":
            df_ = self.get_data()
            df = df_.drop_duplicates("Drug", keep=False).reset_index(drop=True)

        self.entity1 = df.Drug.values
        self.y = df.Y.values
        self.entity1_idx = df.Drug_ID.values
        return df



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT
import os

from .base_group import BenchmarkGroup


class ProteinPeptideGroup(BenchmarkGroup):
    """Create Protein-Peptide Group Class object. This is for benchmarking models predicting protein-peptide interactions.

    Args:
            path (str, optional): the path to store/retrieve the Protein-Peptide group datasets.
    """

    def __init__(self, path="./data", file_format="csv"):
        """Create an SCDTI benchmark group class."""
        # super().__init__(name="SCDTI_Group", path=path)
        self.name = "ProteinPeptide_Group"
        self.path = os.path.join(path, self.name)
        # self.datasets = ["opentargets_dti"]
        self.dataset_names = ["brown_mdm2_ace2_12ca5"]
        self.file_format = file_format
        self.split = None

    def get_train_valid_split(self):
        """parameters included for compatibility. this benchmark has a fixed train/test split."""
        import pandas as pd
        from sklearn.model_selection import train_test_split
        from ..multi_pred.proteinpeptide import ProteinPeptide as DataLoader
        if self.split is None:
            dl = DataLoader(name="brown_mdm2_ace2_12ca5")
            df = dl.get_data()
            for idx, e in enumerate(df["Y"]):
                if e != "Putative binder":
                    df["Y"][idx] = "1"
                else:
                    df["Y"][idx] = "0"
            # raise Exception("unique", )
            # Split the data while stratifying
            X_train, X_test, y_train, y_test = train_test_split(
                df.drop('Y', axis=1),  # features
                df['Y'],  # labels
                test_size=0.9,  # 90% of the data goes to the test set
                random_state=42,  # for reproducibility
                stratify=df[
                    'Y']  # stratify by the label column to ensure even distribution
            )
            self.split = {}
            self.split["train"] = (X_train, y_train)
            self.split["test"] = (X_test, y_test)
            self.split["dev"] = []

        return self.split["train"], self.split["dev"]

    def get_test(self):
        from ..multi_pred.proteinpeptide import ProteinPeptide as DataLoader
        if self.split is None:
            self.get_train_valid_split()
        return self.split["test"]

    def evaluate(self, y_pred):
        from sklearn.metrics import precision_score, recall_score, accuracy_score, f1_score, roc_auc_score
        y_true = self.get_test()[1]
        # Calculate metrics
        precision = precision_score(y_true, y_pred, pos_label="1")
        recall = recall_score(y_true, y_pred, pos_label="1")
        accuracy = accuracy_score(y_true, y_pred)
        f1 = f1_score(y_true, y_pred, pos_label="1")
        auc = roc_auc_score(y_true, y_pred)
        return [precision, recall, accuracy, f1]

    def evaluate_many(self, preds):
        from numpy import mean, std
        if len(preds) < 5:
            raise Exception(
                "Run your model on at least 5 seeds to compare results and provide your outputs in preds."
            )
        out = dict()
        preds = [self.evaluate(p) for p in preds]
        out["precision"] = (mean([x[0] for x in preds]),
                            std([x[0] for x in preds]))
        out["recall"] = (mean([x[1] for x in preds]), std([x[1] for x in preds
                                                          ]))
        out["accuracy"] = (mean([x[2] for x in preds]),
                           std([x[2] for x in preds]))
        out["f1"] = (mean([x[3] for x in preds]), std([x[3] for x in preds]))
        # out["auc"] = (mean([x[4] for x in preds]), std([x[4] for x in preds]))
        return out



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

from .base_group import BenchmarkGroup


class drugcombo_group(BenchmarkGroup):
    """create a drug combination benchmark group

    Args:
        path (str, optional): path to save/load benchmarks
    """

    def __init__(self, path="./data"):
        """create a drug combination benchmark group"""
        super().__init__(name="DrugCombo_Group", path=path, file_format="pkl")

    def get_cell_line_meta_data(self):
        import os
        from ..utils.load import download_wrapper
        from ..utils import load_dict
        name = download_wrapper('drug_comb_meta_data', self.path,
                                ['drug_comb_meta_data'])
        return load_dict(os.path.join(self.path, name + '.pkl'))



from .evaluator import Evaluator
from .oracles import Oracle
from .benchmark_deprecated import BenchmarkGroup
from .model_server.tdc_hf import tdc_hf_interface
from tdc.utils.knowledge_graph import KnowledgeGraph


from .multi_pred_dataset import DataLoader as DL
from ..dataset_configs.config_map import ConfigMap
from ..feature_generators.anndata_to_dataframe import AnnDataToDataFrame


class DataLoader(DL):

    def __init__(self,
                 name,
                 path,
                 print_stats=False,
                 dataset_names=None,
                 no_convert=False):
        super(DataLoader, self).__init__(name, path, print_stats, dataset_names)
        self.adata = self.df  # this is in AnnData format
        cmap = ConfigMap()
        self.cmap = cmap
        self.config = cmap.get(name)
        if no_convert:
            return
        if self.config is None:
            # default to converting adata to dataframe as is
            self.df = AnnDataToDataFrame.anndata_to_df(self.adata)
        else:
            cf = self.config()
            self.df = cf.processing_callback(self.adata)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from ..utils.load import download_wrapper, pd_load
from . import bi_pred_dataset, multi_pred_dataset
from ..metadata import dataset_names


class DrugRes(bi_pred_dataset.DataLoader):
    """Data loader class to load datasets in Drug Response Prediction Task.
    More info: https://tdcommons.ai/multi_pred_tasks/drugres/

    Task Description: Regression. Given the gene expression of cell lines and the SMILES of drug, predict the drug sensitivity level.

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False

    """

    def __init__(self, name, path="./data", label_name=None, print_stats=False):
        """Create Drug Response Prediction dataloader object"""
        super().__init__(name,
                         path,
                         label_name,
                         print_stats,
                         dataset_names=dataset_names["DrugRes"])
        self.entity1_name = "Drug"
        self.entity2_name = "Cell Line"
        self.two_types = True
        self.path = path

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)

    def get_gene_symbols(self):
        """
        Retrieve the gene symbols for the cell line gene expression
        """
        path = self.path
        name = download_wrapper("gdsc_gene_symbols", path,
                                ["gdsc_gene_symbols"])
        print_sys("Loading...")
        import pandas as pd
        import os

        df = pd.read_csv(os.path.join(path, name + ".tab"), sep="\t")
        return df.values.reshape(-1,)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from . import bi_pred_dataset, multi_pred_dataset
from ..metadata import dataset_names


class DTI(bi_pred_dataset.DataLoader):
    """Data loader class to load datasets in Drug-Target Interaction Prediction task.
    More info: https://tdcommons.ai/multi_pred_tasks/dti/

    Regression task. Given the target amino acid sequence/compound SMILES string, predict their binding affinity.


    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False


    """

    def __init__(self, name, path="./data", label_name=None, print_stats=False):
        """Create Drug-Target Interaction Prediction dataloader object"""
        super().__init__(name,
                         path,
                         label_name,
                         print_stats,
                         dataset_names=dataset_names["DTI"])
        self.entity1_name = "Drug"
        self.entity2_name = "Target"
        self.two_types = True

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)

    def harmonize_affinities(self, mode=None):
        """Removing duplicated drug-target pairs with different binding affinities."""

        if mode not in ["mean", "max_affinity"]:
            raise ValueError(
                "Please specify 'mode' of removal, currently supported 'mean'/'max_affinity'!"
            )

        if mode == "max_affinity":
            df_ = self.get_data()
            if self.log_flag:
                print_sys(
                    "The scale is converted to log scale, so we will take the maximum!"
                )
                df = (df_.groupby(["Drug_ID", "Drug", "Target_ID",
                                   "Target"]).Y.agg(max).reset_index())
            else:
                print_sys(
                    "The scale is in original affinity scale, so we will take the minimum!"
                )
                df = (df_.groupby(["Drug_ID", "Drug", "Target_ID",
                                   "Target"]).Y.agg(min).reset_index())

        elif mode == "mean":
            import numpy as np

            df_ = self.get_data()
            df = (df_.groupby(["Drug_ID", "Drug", "Target_ID",
                               "Target"]).Y.agg(np.mean).reset_index())

        self.entity1_idx = df.Drug_ID.values
        self.entity2_idx = df.Target_ID.values

        self.entity1 = df.Drug.values
        self.entity2 = df.Target.values
        self.y = df.Y.values
        print_sys("The original data has been updated!")
        return df


# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import sys
import warnings

warnings.filterwarnings("ignore")

from . import single_pred_dataset
from ..utils import print_sys
from ..metadata import dataset_names


class Develop(single_pred_dataset.DataLoader):
    """Data loader class to load datasets in Develop task. More info: https://tdcommons.ai/single_pred_tasks/develop/

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False
        convert_format (str, optional):
            Automatic conversion of SMILES to other molecular formats in MolConvert class. Stored as separate column in dataframe, defaults to None
    """

    def __init__(
        self,
        name,
        path="./data",
        label_name=None,
        print_stats=False,
        convert_format=None,
    ):
        """Create a Developability prediction dataloader object."""
        super().__init__(
            name,
            path,
            label_name,
            print_stats,
            dataset_names=dataset_names["Develop"],
            convert_format=convert_format,
        )
        self.entity1_name = "Antibody"
        if print_stats:
            self.print_stats()
        print("Done!", flush=True, file=sys.stderr)

    def graphein(
        self,
        graph="distance",
        node_feature=["amino_acid_one_hot"],
        distance_threshold=6,
        config=None,
        convertor=None,
    ):
        from typing import Dict
        import torch
        from functools import partial

        import graphein.protein as gp
        from graphein.ml.conversion import GraphFormatConvertor
        from graphein.ml import InMemoryProteinGraphDataset, ProteinGraphDataset

        from graphein.protein.utils import get_obsolete_mapping
        import warnings

        warnings.filterwarnings("ignore")
        import logging

        logging.getLogger("graphein").setLevel("ERROR")
        try:
            split = self.split
        except:
            raise ValueError("Please define data.split() first!")

        print_sys(
            "Note that this task has obsolete PDBs, thus, we remove it in default. For fair comparison using other models, please retrieve the data.split object for the post removal splits!"
        )

        from graphein.protein.utils import get_obsolete_mapping

        obs = get_obsolete_mapping()
        train_obs = [
            t for t in split["train"]["Antibody_ID"] if t in obs.keys()
        ]
        valid_obs = [
            t for t in split["valid"]["Antibody_ID"] if t in obs.keys()
        ]
        test_obs = [t for t in split["test"]["Antibody_ID"] if t in obs.keys()]

        split["train"] = split["train"].loc[~split["train"]["Antibody_ID"].
                                            isin(train_obs)]
        split["test"] = split["test"].loc[~split["test"]["Antibody_ID"].
                                          isin(test_obs)]
        split["valid"] = split["valid"].loc[~split["valid"]["Antibody_ID"].
                                            isin(valid_obs)]

        self.split = split

        def get_label_map(split_name: str) -> Dict[str, torch.Tensor]:
            return dict(
                zip(
                    split[split_name].Antibody_ID,
                    split[split_name].Y.apply(torch.tensor),
                ))

        train_labels = get_label_map("train")
        valid_labels = get_label_map("valid")
        test_labels = get_label_map("test")

        if config is None:
            if graph == "distance":
                edge_fct = [
                    partial(
                        gp.add_distance_threshold,
                        threshold=distance_threshold,
                        long_interaction_threshold=0,
                    )
                ]

            node_feature_fct = []
            for i in node_feature:
                if i == "amino_acid_one_hot":
                    node_feature_fct.append(gp.amino_acid_one_hot)

            from functools import partial

            graphein_config = gp.ProteinGraphConfig(
                node_metadata_functions=node_feature_fct,
                edge_construction_functions=edge_fct,
            )

            convertor = GraphFormatConvertor(
                src_format="nx",
                dst_format="pyg",
                columns=["coords", "edge_index"] + node_feature,
            )
        else:
            graphein_config = config
            convertor = convertor

        train_ds = InMemoryProteinGraphDataset(
            root="./data/",
            name="train",
            pdb_codes=split["train"]["Antibody_ID"],
            graph_label_map=train_labels,
            graphein_config=graphein_config,
            graph_format_convertor=convertor,
        )

        valid_ds = InMemoryProteinGraphDataset(
            root="./data/",
            name="valid",
            pdb_codes=split["valid"]["Antibody_ID"],
            graph_label_map=valid_labels,
            graphein_config=graphein_config,
            graph_format_convertor=convertor,
        )

        test_ds = InMemoryProteinGraphDataset(
            root="./data/",
            name="test",
            pdb_codes=split["test"]["Antibody_ID"],
            graph_label_map=test_labels,
            graphein_config=graphein_config,
            graph_format_convertor=convertor,
        )

        return {"train": train_ds, "valid": valid_ds, "test": test_ds}



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT
import numpy as np
import os
import pandas as pd

from .base_group import BenchmarkGroup
from ..dataset_configs.config_map import scperturb_datasets, scperturb_gene_datasets


class CounterfactualGroup(BenchmarkGroup):
    """Create Counterfactual Group Class object. This is for single-cell counterfactual prediction tasks (drug, gene) benchmark.

    Args:
            path (str, optional): the path to store/retrieve the Counterfactual group datasets.
    """
    _DRUG_COLS = [
        "ncounts", 'celltype', 'cell_line', 'cancer', 'disease', 'tissue_type',
        'perturbation', 'perturbation_type', 'ngenes'
    ]

    _GENE_COLS = [
        'UMI_count', 'cancer', 'cell_line', 'disease', 'guide_id', 'ncounts',
        'ngenes', 'nperts', 'organism', 'percent_mito', 'percent_ribo',
        'perturbation', 'perturbation_type', 'tissue_type'
    ]

    def __init__(self, path="./data", file_format="csv", is_drug=True):
        """Create a Counterfactual prediction benchmark group class."""
        self.name = "Coutnerfactual_Group"
        self.path = os.path.join(path, self.name)
        self.is_drug = is_drug
        self.dataset_names = scperturb_gene_datasets if not self.is_drug else scperturb_datasets
        self.file_format = file_format
        self.split = None

    def get_train_valid_split(self,
                              dataset=None,
                              split_to_unseen=False,
                              remove_unseen=True):
        """parameters included for compatibility. this benchmark has a fixed train/test split."""
        from ..multi_pred.perturboutcome import PerturbOutcome
        dataset = dataset or "scperturb_drug_AissaBenevolenskaya2021"
        assert dataset in self.dataset_names, "{} dataset not in {}".format(
            dataset, self.dataset_names)
        data = PerturbOutcome(dataset)
        self.split = data.get_split(unseen=split_to_unseen,
                                    remove_unseen=remove_unseen)
        cell_lines = list(self.split.keys())
        self.split["adj"] = 0
        for line in cell_lines:
            print("processing benchmark line", line)
            for split, df in self.split[line].items():
                if split not in self.split:
                    self.split[split] = {}
                elif split == "adj":
                    self.split["adj"] += df
                    continue
                self.split[split][line] = df
            print("done with line", line)
        return self.split["train"], self.split["dev"]

    def get_test(self):
        if self.split is None:
            self.get_train_valid_split()
        return self.split["test"]

    def evaluate(self, y_pred):
        from sklearn.metrics import r2_score
        y_true = self.get_test()
        r2vec = []
        for cell_line, df in y_true.items():
            check = self._DRUG_COLS[0] if self.is_drug else self._GENE_COLS[0]
            cols = self._DRUG_COLS if self.is_drug else self._GENE_COLS
            if check in df.columns:
                df.drop(cols, axis=1)
            if check in y_pred[cell_line].columns:
                y_pred[cell_line].drop(cols, axis=1)
            categorical_cols = df.select_dtypes(
                include=['object', 'category']).columns
            df = df.drop(columns=categorical_cols)
            categorical_cols = y_pred[cell_line].select_dtypes(
                include=['object', 'category']).columns
            y_pred[cell_line] = y_pred[cell_line].drop(columns=categorical_cols)
            mdf = df.mean()
            mpred = y_pred[cell_line].mean()
            if len(mdf) != len(mpred):
                raise Exception(
                    "lengths between true and test mean vectors defers in cell line {} with {} vs {}"
                    .format(cell_line, len(mdf), len(mpred)))
            elif pd.isna(mdf.values).any():
                raise Exception(
                    "ground truth mean contains {} nan values".format(
                        mdf.isna().sum()))
            elif pd.isna(mpred.values).any():
                raise Exception("prediction mean contains {} nan values".format(
                    mpred.isna().sum()))
            r2 = r2_score(mdf, mpred)
            r2vec.append(r2)
        return np.mean(r2vec)

    def evaluate_many(self, preds):
        from numpy import mean, std
        if len(preds) < 5:
            raise Exception(
                "Run your model on at least 5 seeds to compare results and provide your outputs in preds."
            )
        out = dict()
        preds = [self.evaluate(p) for p in preds]
        out["mean_R^2"] = mean(preds)
        out["std_R^2"] = std(preds)
        return out



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .. import base_dataset
from ..utils import (
    dataset2target_lists,
    property_dataset_load,
    create_fold,
    create_fold_setting_cold,
    create_scaffold_split,
    print_sys,
)


class DataLoader(base_dataset.DataLoader):
    """A base data loader class.

    Args:
        name (str): the dataset name.
        path (str): The path to save the data file
        label_name (str): For multi-label dataset, specify the label name
        print_stats (bool): Whether to print basic statistics of the dataset
        dataset_names (list): A list of dataset names available for a task
        convert_format (str): Automatic conversion of SMILES to other molecular formats in MolConvert class. Stored as separate column in dataframe

    Attributes:
        convert_format (str): conversion format of an entity
        convert_result (list): a placeholder for a list of conversion outputs
        entity1 (Pandas Series): a list of the single entites
        entity1_idx (Pandas Series): a list of the single entites index
        entity1_name (Pandas Series): a list of the single entites names
        file_format (str): the format of the downloaded dataset
        label_name (str): for multi-label dataset, the label name of interest
        name (str): dataset name
        path (str): path to save and retrieve the dataset
        y (Pandas Series): a list of the single entities label
    """

    def __init__(
        self,
        name,
        path,
        label_name,
        print_stats,
        dataset_names,
        convert_format,
        raw_format="SMILES",
    ):
        """Create a base dataloader object that each single instance prediction task dataloader class can inherit from.

        Raises:
            ValueError: for a dataset with multiple labels, specify the label. Use tdc.utils.retrieve_label_name_list to see the available label names

        """
        if name.lower() in dataset2target_lists.keys():
            if label_name is None:
                raise ValueError(
                    "Please select a label name. You can use tdc.utils.retrieve_label_name_list('"
                    + name.lower() +
                    "') to retrieve all available label names.")

        entity1, y, entity1_idx = property_dataset_load(name, path, label_name,
                                                        dataset_names)

        self.entity1 = entity1
        self.y = y
        self.entity1_idx = entity1_idx
        self.name = name
        self.entity1_name = "Drug"
        self.path = path
        self.file_format = "csv"
        self.label_name = label_name
        self.convert_format = convert_format
        self.convert_result = None
        self.raw_format = raw_format  ### 'SMILES' for most data, 'Raw3D' for QM9, ...

    def get_data(self, format="df"):
        """
        Arguments:
            format (str, optional): the returning dataset format, defaults to 'df'

        Returns:
            pandas DataFrame/dict: a dataframe of a dataset/a dictionary for key information in the dataset

        Raises:
            AttributeError: Use the correct format input (df, dict, DeepPurpose)
        """

        if (self.convert_format is not None) and (self.convert_result is None):
            from ..chem_utils import MolConvert

            converter = MolConvert(src=self.raw_format, dst=self.convert_format)
            convert_result = converter(self.entity1.values)
            self.convert_result = [i for i in convert_result]

        if format == "df":
            if self.convert_format is not None:
                return pd.DataFrame({
                    self.entity1_name + "_ID":
                        self.entity1_idx,
                    self.entity1_name:
                        self.entity1,
                    self.entity1_name + "_" + self.convert_format:
                        self.convert_result,
                    "Y":
                        self.y,
                })
            else:
                return pd.DataFrame({
                    self.entity1_name + "_ID": self.entity1_idx,
                    self.entity1_name: self.entity1,
                    "Y": self.y,
                })

        elif format == "dict":
            if self.convert_format is not None:
                return {
                    self.entity1_name + "_ID":
                        self.entity1_idx.values,
                    self.entity1_name:
                        self.entity1.values,
                    self.entity1_name + "_" + self.convert_format:
                        self.convert_result,
                    "Y":
                        self.y.values,
                }
            else:
                return {
                    self.entity1_name + "_ID": self.entity1_idx.values,
                    self.entity1_name: self.entity1.values,
                    "Y": self.y.values,
                }
        elif format == "DeepPurpose":
            return self.entity1.values, self.y.values
        else:
            raise AttributeError("Please use the correct format input")

    def get_split(self, method="random", seed=42, frac=[0.7, 0.1, 0.2]):
        """
        Arguments:
            method: splitting schemes, choose from random, cold_{entity}, scaffold, defaults to 'random'
            seed: the random seed for splitting dataset, defaults to '42'
            frac: train/val/test split fractions, defaults to '[0.7, 0.1, 0.2]'

        Returns:
            dict: a dictionary with three keys ('train', 'valid', 'test'), each value is a pandas dataframe object of the splitted dataset

        Raises:
            AttributeError: the input split method is not available.
        """

        df = self.get_data(format="df")

        if method == "random":
            split = create_fold(df, seed, frac)
        elif method == "cold_" + self.entity1_name.lower():
            split = create_fold_setting_cold(df, seed, frac, self.entity1_name)
        elif method == "scaffold":
            split = create_scaffold_split(df, seed, frac, self.entity1_name)
        else:
            raise AttributeError("Please specify the correct splitting method")
        self.split = split
        return split

    def print_stats(self):
        """Print basic data statistics."""
        print_sys("--- Dataset Statistics ---")
        try:
            x = np.unique(self.entity1)
        except:
            x = np.unique(self.entity1_idx)

        print(
            str(len(x)) + " unique " + self.entity1_name.lower() + "s.",
            flush=True,
            file=sys.stderr,
        )
        print_sys("--------------------------")



"""wrapper for download various dataset 
"""
import requests
from zipfile import ZipFile
import os, sys
import pandas as pd
import numpy as np
import pickle
from pandas.errors import EmptyDataError
from tqdm import tqdm

from ..metadata import (
    name2type,
    name2id,
    name2idlist,
    dataset_list,
    dataset_names,
    benchmark_names,
    benchmark2id,
    benchmark2type,
)
from ..metadata import (
    property_names,
    paired_dataset_names,
    single_molecule_dataset_names,
)
from ..metadata import (
    retrosyn_dataset_names,
    forwardsyn_dataset_names,
    molgenpaired_dataset_names,
    generation_datasets,
)
from ..metadata import (
    oracle2id,
    receptor2id,
    download_oracle_names,
    trivial_oracle_names,
    oracle_names,
    oracle2type,
)
from collections import defaultdict

receptor_names = list(receptor2id.keys())
sys.path.append("../")

from .misc import fuzzy_search, print_sys


def download_wrapper(name, path, dataset_names):
    """wrapper for downloading a dataset given the name and path, for csv,pkl,tsv files

    Args:
        name (str): the rough dataset query name
        path (str): the path to save the dataset
        dataset_names (list): the list of available dataset names to search the query dataset

    Returns:
        str: the exact dataset query name
    """
    name = fuzzy_search(name, dataset_names)
    server_path = "https://dataverse.harvard.edu/api/access/datafile/"

    if name in name2idlist:
        for i, id in enumerate(name2idlist[name]):
            dataset_path = server_path + str(id)

            if not os.path.exists(path):
                os.mkdir(path)

            if os.path.exists(
                    os.path.join(
                        path, name + "-" + str(i + 1) + "." + name2type[name])):
                print_sys("Found local copy...")
            else:
                print_sys("Downloading...")
                dataverse_download(dataset_path,
                                   path,
                                   name,
                                   name2type,
                                   id=i + 1)

        return name
    else:
        dataset_path = server_path + str(name2id[name])

        if not os.path.exists(path):
            os.mkdir(path)

        if os.path.exists(os.path.join(path, name + "." + name2type[name])):
            print_sys("Found local copy...")
        else:
            print_sys("Downloading...")
            dataverse_download(dataset_path, path, name, name2type)

        return name


def zip_data_download_wrapper(name, path, dataset_names):
    """wrapper for downloading a dataset given the name and path - zip file, automatically unzipping

    Args:
        name (str): the rough dataset query name
        path (str): the path to save the dataset
        dataset_names (list): the list of available dataset names to search the query dataset

    Returns:
        str: the exact dataset query name
    """
    name = fuzzy_search(name, dataset_names)
    server_path = "https://dataverse.harvard.edu/api/access/datafile/"

    if name in name2idlist:
        for i, id in enumerate(name2idlist[name]):
            dataset_path = server_path + str(id)

            if not os.path.exists(path):
                os.mkdir(path)

            if os.path.exists(os.path.join(path, name + "-" + str(i + 1))):
                print_sys(
                    f"Found local copy for {i+1}/{len(name2idlist[name])} file..."
                )
            else:
                print_sys(f"Downloading {i+1}/{len(name2idlist[name])} file...")
                dataverse_download(dataset_path,
                                   path,
                                   name,
                                   name2type,
                                   id=i + 1)
                print_sys(
                    f"Extracting zip {i+1}/{len(name2idlist[name])} file...")
                with ZipFile(
                        os.path.join(path, name + "-" + str(i + 1) + ".zip"),
                        "r") as zip:
                    zip.extractall(path=os.path.join(path))
        if not os.path.exists(os.path.join(path, name)):
            os.mkdir(os.path.join(path, name))
        for i in range(len(name2idlist[name])):
            os.system(f"mv {path}/{name}-{i+1}/* {path}/{name} 2>/dev/null")
        print_sys("Done!")
    else:
        dataset_path = server_path + str(name2id[name])

        if not os.path.exists(path):
            os.mkdir(path)

        if os.path.exists(os.path.join(path, name)):
            print_sys("Found local copy...")
        else:
            print_sys("Downloading...")
            dataverse_download(dataset_path, path, name, name2type)
            print_sys("Extracting zip file...")
            with ZipFile(os.path.join(path, name + ".zip"), "r") as zip:
                zip.extractall(path=os.path.join(path))
            print_sys("Done!")
    return name


def dataverse_download(url, path, name, types, id=None):
    """dataverse download helper with progress bar

    Args:
        url (str): the url of the dataset
        path (str): the path to save the dataset
        name (str): the dataset name
        types (dict): a dictionary mapping from the dataset name to the file format
    """
    if id is None:
        save_path = os.path.join(path, name + "." + types[name])
    else:
        save_path = os.path.join(path, name + "-" + str(id) + "." + types[name])
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    with open(save_path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()


def oracle_download_wrapper(name, path, oracle_names):
    """wrapper for downloading an oracle model checkpoint given the name and path

    Args:
        name (str): the rough oracle query name
        path (str): the path to save the oracle
        dataset_names (list): the list of available exact oracle names

    Returns:
        str: the exact oracle query name
    """
    name = fuzzy_search(name, oracle_names)
    if name in trivial_oracle_names:
        return name

    server_path = "https://dataverse.harvard.edu/api/access/datafile/"
    dataset_path = server_path + str(oracle2id[name])

    if not os.path.exists(path):
        os.mkdir(path)

    if os.path.exists(os.path.join(path, name + "." + oracle2type[name])):
        print_sys("Found local copy...")
    else:
        print_sys("Downloading Oracle...")
        dataverse_download(dataset_path, path, name,
                           oracle2type)  ## to-do to-check
        print_sys("Done!")
    return name


def receptor_download_wrapper(name, path):
    """wrapper for downloading an receptor pdb file given the name and path

    Args:
        name (str): the exact pdbid
        path (str): the path to save the oracle

    Returns:
        str: the exact pdbid
    """

    server_path = "https://dataverse.harvard.edu/api/access/datafile/"
    dataset_paths = [
        server_path + str(receptor2id[name][0]),
        server_path + str(receptor2id[name][1]),
    ]

    if not os.path.exists(path):
        os.mkdir(path)

    if os.path.exists(os.path.join(path, name + ".pdbqt")) and os.path.exists(
            os.path.join(path, name + ".pdb")):
        print_sys("Found local copy...")
    else:
        print_sys("Downloading receptor...")
        receptor2type = defaultdict(lambda: "pdbqt")
        dataverse_download(dataset_paths[0], path, name,
                           receptor2type)  ## to-do to-check
        receptor2type = defaultdict(lambda: "pdb")
        dataverse_download(dataset_paths[1], path, name,
                           receptor2type)  ## to-do to-check
        print_sys("Done!")
    return name


def bm_download_wrapper(name, path):
    """wrapper for downloading a benchmark group given the name and path

    Args:
        name (str): the rough benckmark group query name
        path (str): the path to save the benchmark group
        dataset_names (list): the list of available benchmark group names

    Returns:
        str: the exact benchmark group query name
    """
    name = fuzzy_search(name, list(benchmark_names.keys()))
    server_path = "https://dataverse.harvard.edu/api/access/datafile/"
    dataset_path = server_path + str(benchmark2id[name])

    if not os.path.exists(path):
        os.mkdir(path)

    if os.path.exists(os.path.join(path, name)):
        print_sys("Found local copy...")
    else:
        print_sys("Downloading Benchmark Group...")
        dataverse_download(dataset_path, path, name, benchmark2type)
        print_sys("Extracting zip file...")
        with ZipFile(os.path.join(path, name + ".zip"), "r") as zip:
            zip.extractall(path=os.path.join(path))
        print_sys("Done!")
    return name


def pd_load(name, path):
    """load a pandas dataframe from local file.

    Args:
        name (str): dataset name
        path (str): the path where the dataset is saved

    Returns:
        pandas.DataFrame: loaded dataset in dataframe

    Raises:
        ValueError: the file format is not supported. currently only support tab/csv/pkl/zip
    """
    try:
        if name2type[name] == "tab":
            df = pd.read_csv(os.path.join(path, name + "." + name2type[name]),
                             sep="\t")
        elif name2type[name] == "csv":
            df = pd.read_csv(os.path.join(path, name + "." + name2type[name]))
        elif name2type[name] == "xlsx":
            df = pd.read_excel(os.path.join(path, name + "." + name2type[name]))
        elif name2type[name] == "pkl":
            df = pd.read_pickle(os.path.join(path,
                                             name + "." + name2type[name]))
        elif name2type[name] == "zip":
            df = pd.read_pickle(os.path.join(path, name + "/" + name + ".pkl"))
        elif name2type[name] == "h5ad":
            import anndata
            adata = anndata.read_h5ad(
                os.path.join(path, name + "." + name2type[name]))
            # df = pd.DataFrame(adata.X.toarray(), columns=adata.var_names, index=adata.obs_names)
            # TODO: multi-index would help include var information in columns
            # multi_index = pd.MultiIndex.from_frame(adata.var.reset_index())
            # df.columns = multi_index
            return adata
        elif name2type[name] == "json":
            # df = pd.read_json(os.path.join(path, name + "." + name2type[name]))
            import json
            file_path = os.path.join(path, name + "." + name2type[name])
            with open(file_path, 'r') as f:
                file_content = json.load(f)
            try:
                maxlen = max(len(x) for x in file_content.values())
            except:
                return file_content
            for k, v in file_content.items():
                r = maxlen - len(v)
                file_content[k] = v + [None] * r
            df = pd.DataFrame(file_content)

        elif name2type[name] == "pth":
            import torch
            tensors = torch.load(
                os.path.join(path, name + "." + name2type[name]))
            dfs = {}
            if isinstance(tensors, dict):
                for k, v in tensors.items():
                    if isinstance(v, torch.Tensor):
                        dfs[k] = pd.DataFrame(v.detach().numpy())
                    else:
                        raise Exception("encountered non-tensor")
                df = pd.concat(dfs, axis=0)
            elif torch.is_tensor(tensors):
                df = pd.DataFrame(tensors.detach().numpy())
            else:
                raise Exception("encountered non-tensor")

        else:
            raise ValueError(
                "The file type must be one of tab/csv/xlsx/pickle/zip.")
        try:
            df = df.drop_duplicates()
        except:
            pass
        return df
    except (EmptyDataError, EOFError) as e:
        import sys

        sys.exit(
            "TDC is hosted in Harvard Dataverse and it is currently under maintenance, please check back in a few hours or checkout https://dataverse.harvard.edu/."
        )


def load_json_from_txt_file(name, path):
    import json
    import re
    name = download_wrapper(name, path, [name])
    file_path = os.path.join(path, name + ".txt")
    with open(file_path, 'r') as f:
        data = f.read()
        # data = re.sub(r"(?<!\\)'", '"', data)
        data = data.replace("\'", "\"")
        file_content = json.loads(data)
    maxlen = max(len(x) for x in file_content.values())
    for k, v in file_content.items():
        r = maxlen - len(v)
        file_content[k] = v + [None] * r
    df = pd.DataFrame(file_content)
    return df


def property_dataset_load(name, path, target, dataset_names):
    """a wrapper to download, process and load single-instance prediction task datasets

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        target (str): for multi-label dataset, retrieve the label of interest
        dataset_names (list): a list of availabel exact dataset names

    Returns:
        pandas.Series: three series (entity representation, label, entity id)
    """
    if target is None:
        target = "Y"
    name = download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    try:
        if target is not None:
            target = fuzzy_search(target, df.columns.values)
        # df = df.T.drop_duplicates().T ### does not work
        # df2 = df.loc[:,~df.T.duplicated(keep='first')]  ### does not work
        df2 = df.loc[:,
                     ~df.columns.duplicated()]  ### remove the duplicate columns
        df = df2
        df = df[df[target].notnull()].reset_index(drop=True)
    except:
        with open(os.path.join(path, name + "." + name2type[name]), "r") as f:
            if name2type[name] == "pkl":
                import pickle

                file_content = pickle.load(
                    open(os.path.join(path, name + "." + name2type[name]),
                         "rb"))
            else:
                file_content = " ".join(f.readlines())
            flag = "Service Unavailable" in " ".join(file_content)
            # flag = 'Service Unavailable' in ' '.join(f.readlines())
            if flag:
                import sys

                sys.exit(
                    "TDC is hosted in Harvard Dataverse and it is currently under maintenance, please check back in a few hours or checkout https://dataverse.harvard.edu/."
                )
            else:
                import sys

                sys.exit(
                    "Please report this error to contact@tdcommons.ai, thanks!")
    try:
        return df["X"], df[target], df["ID"]
    except:
        return df["Drug"], df[target], df["Drug_ID"]


def resource_dataset_load(name, path, dataset_names):
    if name not in dataset_names:
        raise ValueError(
            "Unknown resource dataset {}, should be one of: {}".format(
                name, dataset_names))
    name = download_wrapper(name, path, dataset_names)
    return pd_load(name, path)


def interaction_dataset_load(name,
                             path,
                             target,
                             dataset_names,
                             aux_column,
                             data_config=None):
    """a wrapper to download, process and load two-instance prediction task datasets

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        target (str): for multi-label dataset, retrieve the label of interest
        dataset_names (list): a list of availabel exact dataset names
        var_map (dict): maps variable names X1, X2, ID1, ID2 into existing column names

    Returns:
        pandas.Series: three series (entity 1 representation, entity 2 representation, entity id 1, entity id 2, label)
    """
    name = download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    if data_config is not None:
        # code block to apply preprocessing rules defined by config files
        process_callback = data_config.processing_callback
        if process_callback is not None:
            df = process_callback(df)
        tdc_standard_callback = data_config.tdc_cols_callback
        if tdc_standard_callback is not None:
            df = tdc_standard_callback(df)
    try:
        if target is None:
            target = "Y"
        if target not in df.columns.values:
            # for binary interaction data, the labels are all 1. negative samples can be sampled from utils.NegSample function
            df[target] = 1
        if target is not None:
            target = fuzzy_search(target, df.columns.values)
        df = df[df[target].notnull()].reset_index(drop=True)
        if aux_column is None:
            return df["X1"], df["X2"], df[target], df["ID1"], df[
                "ID2"], "_", df, data_config is not None
        else:
            return df["X1"], df["X2"], df[target], df["ID1"], df["ID2"], df[
                aux_column], df, data_config is not None

    except:
        with open(os.path.join(path, name + "." + name2type[name]), "r") as f:
            flag = "Service Unavailable" in " ".join(f.readlines())
            if flag:
                import sys

                sys.exit(
                    "TDC is hosted in Harvard Dataverse and it is currently under maintenance, please check back in a few hours or checkout https://dataverse.harvard.edu/."
                )
            else:
                import sys

                sys.exit(
                    "Please report this error to cosamhkx@gmail.com, thanks!")


def multi_dataset_load(name, path, dataset_names):
    """a wrapper to download, process and load multiple(>2)-instance prediction task datasets. assume the downloaded file is already processed

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        target (str): for multi-label dataset, retrieve the label of interest
        dataset_names (list): a list of availabel exact dataset names

    Returns:
        pandas.DataFrame: the raw dataframe
    """
    name = download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    return df


def generation_paired_dataset_load(name, path, dataset_names, input_name,
                                   output_name):
    """a wrapper to download, process and load generation-paired task datasets

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        target (str): for multi-label dataset, retrieve the label of interest
        dataset_names (list): a list of availabel exact dataset names

    Returns:
        pandas.Series: two series (entity 1 representation, label)
    """
    name = download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    return df[input_name], df[output_name]


def three_dim_dataset_load(name, path, dataset_names):
    """a wrapper to download, process and load 3d molecule task datasets

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        dataset_names (list): a list of availabel exact dataset names

    Returns:
        pandas.DataFrame: the dataframe holds 3d information
        str: the path of the dataset
        str: the name of the dataset
    """
    name = zip_data_download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    return df, os.path.join(path, name), name


def distribution_dataset_load(name, path, dataset_names, column_name):
    """a wrapper to download, process and load molecule distribution learning task datasets. assume the downloaded file is already processed

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        dataset_names (list): a list of availabel exact dataset names
        column_name (str): the column specifying where molecule locates

    Returns:
        pandas.Series: the input list of molecules representation
    """
    name = download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    return df[column_name]


def bi_distribution_dataset_load(
    name,
    path,
    dataset_names,
    return_pocket=False,
    threshold=15,
    remove_protein_Hs=True,
    remove_ligand_Hs=True,
    keep_het=False,
):
    """a wrapper to download, process and load protein-ligand conditional generation task datasets. assume the downloaded file is already processed

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        dataset_names (list): a list of availabel exact dataset names

    Returns:
        pandas.Series: the input list of molecules representation
    """
    name = fuzzy_search(name, dataset_names)
    if name in name2id or name in name2idlist:
        name = zip_data_download_wrapper(name, path, dataset_names)

    if name == "pdbbind":
        print_sys("Processing (this may take long)...")
        protein, ligand = process_pdbbind(path, name, return_pocket,
                                          remove_protein_Hs, remove_ligand_Hs,
                                          keep_het)
    elif name == "dude":
        print_sys("Processing (this may take long)...")
        if return_pocket:
            raise ImportError("DUD-E does not support pocket extraction yet")
        protein, ligand = process_dude(path, name, return_pocket,
                                       remove_protein_Hs, remove_ligand_Hs,
                                       keep_het)
    elif name == "scpdb":
        print_sys("Processing (this may take long)...")
        protein, ligand = process_scpdb(path, name, return_pocket,
                                        remove_protein_Hs, remove_ligand_Hs,
                                        keep_het)
    elif name == "crossdock":
        print_sys("Processing (this may take long)...")
        protein, ligand = process_crossdock(path, name, return_pocket,
                                            remove_protein_Hs, remove_ligand_Hs,
                                            keep_het)

    return protein, ligand


def generation_dataset_load(name, path, dataset_names):
    """a wrapper to download, process and load generation task datasets. assume the downloaded file is already processed

    Args:
        name (str): the rough dataset name
        path (str): the dataset path to save/retrieve
        dataset_names (list): a list of availabel exact dataset names

    Returns:
        pandas.Series: the data series
    """
    name = download_wrapper(name, path, dataset_names)
    print_sys("Loading...")
    df = pd_load(name, path)
    return df["input"], df["target"]


def oracle_load(name, path="./oracle", oracle_names=oracle_names):
    """a wrapper to download, process and load oracles.

    Args:
        name (str): the rough oracle name
        path (str): the oracle path to save/retrieve, defaults to './oracle'
        dataset_names (list): a list of availabel exact oracle names

    Returns:
        str: exact oracle name
    """
    name = oracle_download_wrapper(name, path, oracle_names)
    return name


def receptor_load(name, path="./oracle"):
    """a wrapper to download, process and load pdb file.

    Args:
        name (str): the rough pdbid name
        path (str): the oracle path to save/retrieve, defaults to './oracle'

    Returns:
        str: exact pdbid name
    """
    name = receptor_download_wrapper(name, path)
    return name


def bm_group_load(name, path):
    """a wrapper to download, process and load benchmark group

    Args:
        name (str): the rough benchmark group name
        path (str): the benchmark group path to save/retrieve

    Returns:
        str: exact benchmark group name
    """
    name = bm_download_wrapper(name, path)
    return name


def process_pdbbind(
    path,
    name="pdbbind",
    return_pocket=False,
    threshold=15,
    remove_protein_Hs=True,
    remove_ligand_Hs=True,
    keep_het=False,
):
    """a processor to process pdbbind dataset

    Args:
        name (str): the name of the dataset
        path (str): the path to save the data file
        print_stats (bool): whether to print the basic statistics of the dataset
        return_pocket (bool): whether to return only protein pocket or full protein
            threshold (int): only enabled when return_pocket is to True, if pockets are not provided in the raw data,
                                             the threshold is used as a radius for a sphere around the ligand center to consider protein pocket
            remove_protein_Hs (bool): whether to remove H atoms from proteins or not
            remove_ligand_Hs (bool): whether to remove H atoms from ligands or not
            keep_het (bool): whether to keep het atoms (e.g. cofactors) in protein

    Returns:
            protein (dict): a dict of protein features
            ligand (dict): a dict of ligand features
    """
    from rdkit import Chem, RDLogger

    RDLogger.DisableLog("rdApp.*")
    from biopandas.pdb import PandasPdb

    if os.path.exists(path):
        print_sys("Processing...")
        protein_coords, protein_atom_types = [], []
        ligand_coords, ligand_atom_types = [], []
        files = os.listdir(path)
        failure = 0
        total_ct = 0
        for idx, file in enumerate(tqdm(files)):
            if file == "readme" or file == "index":
                continue
            total_ct += 1
            try:
                if return_pocket:
                    protein = PandasPdb().read_pdb(
                        os.path.join(path, f"{file}/{file}_pocket.pdb"))
                else:
                    protein = PandasPdb().read_pdb(
                        os.path.join(path, f"{file}/{file}_protein.pdb"))
                ligand = Chem.SDMolSupplier(os.path.join(
                    path, f"{file}/{file}_ligand.sdf"),
                                            sanitize=False)[0]
                ligand = extract_atom_from_mol(ligand, remove_ligand_Hs)
                # if ligand contains unallowed atoms
                if ligand is None:
                    continue
                else:
                    ligand_coord, ligand_atom_type = ligand
                protein_coord, protein_atom_type = extract_atom_from_protein(
                    protein.df["ATOM"],
                    protein.df["HETATM"],
                    remove_protein_Hs,
                    keep_het,
                )
                protein_coords.append(protein_coord)
                ligand_coords.append(ligand_coord)
                protein_atom_types.append(protein_atom_type)
                ligand_atom_types.append(ligand_atom_type)
            except:
                failure += 1
                continue
        print_sys(f"processing done, {failure}/{total_ct} fails")
        protein = {"coord": protein_coords, "atom_type": protein_atom_types}
        ligand = {"coord": ligand_coords, "atom_type": ligand_atom_types}
    else:
        sys.exit("Wrong path!")
    return protein, ligand


def process_crossdock(
    path,
    name="crossdock",
    return_pocket=False,
    threshold=15,
    remove_protein_Hs=True,
    remove_ligand_Hs=True,
    keep_het=False,
):
    """a processor to process crossdock dataset

    Args:
        name (str): the name of the dataset
        path (str): the path to save the data file
        print_stats (bool): whether to print the basic statistics of the dataset
        return_pocket (bool): whether to return only protein pocket or full protein
            threshold (int): only enabled when return_pocket is to True, if pockets are not provided in the raw data,
                                             the threshold is used as a radius for a sphere around the ligand center to consider protein pocket
            remove_protein_Hs (bool): whether to remove H atoms from proteins or not
            remove_ligand_Hs (bool): whether to remove H atoms from ligands or not
            keep_het (bool): whether to keep het atoms (e.g. cofactors) in protein

    Returns:
            protein (dict): a dict of protein features
            ligand (dict): a dict of ligand features
    """
    from rdkit import Chem, RDLogger

    RDLogger.DisableLog("rdApp.*")
    from biopandas.pdb import PandasPdb

    protein_coords, protein_atom_types = [], []
    ligand_coords, ligand_atom_types = [], []
    failure = 0
    total_ct = 0
    path = os.path.join(path, name)
    index_path = os.path.join(path, "index.pkl")
    index = pickle.load(open(index_path, "rb"))
    path = os.path.join(path, "crossdocked_pocket10")
    for idx, (pocket_fn, ligand_fn, _, rmsd) in enumerate(tqdm(index)):
        total_ct += 1
        if pocket_fn is None or ligand_fn is None:
            continue
        try:
            if return_pocket:
                protein = PandasPdb().read_pdb(os.path.join(path, pocket_fn))
            else:
                # full protein not stored in the preprocessed crossdock by Luo et al 2021
                protein = PandasPdb().read_pdb(os.path.join(path, pocket_fn))
            ligand = Chem.SDMolSupplier(os.path.join(path, ligand_fn),
                                        sanitize=False)[0]
            ligand = extract_atom_from_mol(ligand, remove_ligand_Hs)
            if ligand is None:
                continue
            else:
                ligand_coord, ligand_atom_type = ligand
            protein_coord, protein_atom_type = extract_atom_from_protein(
                protein.df["ATOM"], protein.df["HETATM"], remove_protein_Hs,
                keep_het)
            protein_coords.append(protein_coord)
            ligand_coords.append(ligand_coord)
            protein_atom_types.append(protein_atom_type)
            ligand_atom_types.append(ligand_atom_type)
        except:
            failure += 1
            continue
    print_sys(f"processing done, {failure}/{total_ct} fails")
    protein = {"coord": protein_coords, "atom_type": protein_atom_types}
    ligand = {"coord": ligand_coords, "atom_type": ligand_atom_types}
    return protein, ligand


def process_dude(
    path,
    name="dude",
    return_pocket=False,
    threshold=15,
    remove_protein_Hs=True,
    remove_ligand_Hs=True,
    keep_het=False,
):
    """a processor to process DUD-E dataset

    Args:
        name (str): the name of the dataset
        path (str): the path to save the data file
        print_stats (bool): whether to print the basic statistics of the dataset
        return_pocket (bool): whether to return only protein pocket or full protein
            threshold (int): only enabled when return_pocket is to True, if pockets are not provided in the raw data,
                                             the threshold is used as a radius for a sphere around the ligand center to consider protein pocket
            remove_protein_Hs (bool): whether to remove H atoms from proteins or not
            remove_ligand_Hs (bool): whether to remove H atoms from ligands or not
            keep_het (bool): whether to keep het atoms (e.g. cofactors) in protein

    Returns:
            protein (dict): a dict of protein features
            ligand (dict): a dict of ligand features
    """
    from rdkit import Chem, RDLogger

    RDLogger.DisableLog("rdApp.*")
    from biopandas.pdb import PandasPdb

    protein_coords, protein_atom_types = [], []
    ligand_coords, ligand_atom_types = [], []
    path = os.path.join(path, name)
    files = os.listdir(path)
    failure = 0
    total_ct = 0
    for idx, file in enumerate(tqdm(files)):
        protein = PandasPdb().read_pdb(
            os.path.join(path, f"{file}/receptor.pdb"))
        if not os.path.exists(os.path.join(path, f"{file}/actives_final.sdf")):
            os.system(f"gzip -d {path}/{file}/actives_final.sdf.gz")
        crystal_ligand = Chem.MolFromMol2File(os.path.join(
            path, f"{file}/crystal_ligand.mol2"),
                                              sanitize=False)
        crystal_ligand = extract_atom_from_mol(crystal_ligand, remove_ligand_Hs)
        if crystal_ligand is None:
            continue
        else:
            crystal_ligand_coord, crystal_ligand_atom_type = crystal_ligand
        ligands = Chem.SDMolSupplier(os.path.join(path,
                                                  f"{file}/actives_final.sdf"),
                                     sanitize=False)
        protein_coord, protein_atom_type = extract_atom_from_protein(
            protein.df["ATOM"], protein.df["HETATM"], remove_protein_Hs,
            keep_het)
        protein_coords.append(protein_coord)
        ligand_coords.append(crystal_ligand_coord)
        protein_atom_types.append(protein_atom_type)
        ligand_atom_types.append(crystal_ligand_atom_type)
        for ligand in ligands:
            total_ct += 1
            try:
                ligand = extract_atom_from_mol(ligand, remove_ligand_Hs)
                # if ligand contains unallowed atoms
                if ligand is None:
                    continue
                else:
                    ligand_coord, ligand_atom_type = ligand
                protein_coords.append(protein_coord)
                ligand_coords.append(ligand_coord)
                protein_atom_types.append(protein_atom_type)
                ligand_atom_types.append(ligand_atom_type)
            except:
                failure += 1
                continue
    print_sys(f"processing done, {failure}/{total_ct} fails")
    protein = {"coord": protein_coords, "atom_type": protein_atom_types}
    ligand = {"coord": ligand_coords, "atom_type": ligand_atom_types}
    return protein, ligand


def process_scpdb(
    path,
    name="scPDB",
    return_pocket=False,
    threshold=15,
    remove_protein_Hs=True,
    remove_ligand_Hs=True,
    keep_het=False,
):
    """a processor to process scpdb dataset

    Args:
        name (str): the name of the dataset
        path (str): the path to save the data file
        print_stats (bool): whether to print the basic statistics of the dataset
        return_pocket (bool): whether to return only protein pocket or full protein
            threshold (int): only enabled when return_pocket is to True, if pockets are not provided in the raw data,
                                             the threshold is used as a radius for a sphere around the ligand center to consider protein pocket
            remove_protein_Hs (bool): whether to remove H atoms from proteins or not
            remove_ligand_Hs (bool): whether to remove H atoms from ligands or not
            keep_het (bool): whether to keep het atoms (e.g. cofactors) in protein

    Returns:
            protein (dict): a dict of protein features
            ligand (dict): a dict of ligand features
    """
    from rdkit import Chem, RDLogger

    RDLogger.DisableLog("rdApp.*")
    from biopandas.mol2 import PandasMol2

    protein_coords, protein_atom_types = [], []
    ligand_coords, ligand_atom_types = [], []
    path = os.path.join(path, name)
    files = os.listdir(path)
    failure = 0
    total_ct = 0
    for idx, file in enumerate(tqdm(files)):
        total_ct += 1
        try:
            if return_pocket:
                protein = PandasMol2().read_mol2(
                    os.path.join(path, f"{file}/site.mol2"))
            else:
                protein = PandasMol2().read_mol2(
                    os.path.join(path, f"{file}/protein.mol2"))
            ligand = Chem.SDMolSupplier(os.path.join(path,
                                                     f"{file}/ligand.sdf"),
                                        sanitize=False)[0]
            ligand = extract_atom_from_mol(ligand, remove_Hs=remove_ligand_Hs)
            # if ligand contains unallowed atoms
            if ligand is None:
                continue
            else:
                ligand_coord, ligand_atom_type = ligand
            protein_coord, protein_atom_type = extract_atom_from_protein(
                protein.df, None, remove_Hs=remove_protein_Hs, keep_het=False)
            protein_coords.append(protein_coord)
            ligand_coords.append(ligand_coord)
            protein_atom_types.append(protein_atom_type)
            ligand_atom_types.append(ligand_atom_type)
        except:
            failure += 1
            continue
    print_sys(f"processing done, {failure}/{total_ct} fails")
    protein = {"coord": protein_coords, "atom_type": protein_atom_types}
    ligand = {"coord": ligand_coords, "atom_type": ligand_atom_types}
    return protein, ligand


def atom_to_one_hot(atom, allowed_atom_list):
    """a helper to convert atom to one-hot encoding

    Args:
            atom (str): the atom to convert
            allowed_atom_list (list(str)): atom types allowed to include

    Returns:
            new_atom (numpy.array): atom one-hot encoding vector
    """
    length = len(allowed_atom_list)
    atom = allowed_atom_list.index(atom)
    new_atom = np.zeros(shape=(length,))
    new_atom[atom] = 1
    return new_atom


def extract_atom_from_mol(rdmol, remove_Hs):
    """a helper to extract molecule atom information

    Args:
            rdmol (rdkit.rdmol): rdkit molecule
            remove_Hs (bool): whether to remove H atoms from ligands or not

    Returns:
            coord (numpy.array): atom types
            atom_type (numpy.array): atom coordinates
    """
    coord = [
        list(rdmol.GetConformer(0).GetAtomPosition(idx))
        for idx in range(rdmol.GetNumAtoms())
        if not remove_Hs or rdmol.GetAtomWithIdx(idx).GetAtomicNum() != 1
    ]
    atom_type = [
        atom.GetAtomicNum()
        for atom in rdmol.GetAtoms()
        if not remove_Hs or atom.GetAtomicNum() != 1
    ]
    return np.array(coord), np.array(atom_type)


def extract_atom_from_protein(data_frame, data_frame_het, remove_Hs, keep_het):
    """a helper to extract protein atom information

    Args:
            data_frame (pandas.dataframe): protein atom
            data_frame_het (pandas.dataframe): protein het atom
            remove_Hs (bool): whether to remove H atoms from proteins or not
            keep_het (bool): whether to keep het atoms (e.g. cofactors) in protein

    Returns:
            coord (numpy.array): atom types
            atom_type (numpy.array): atom coordinates
    """
    from rdkit import RDLogger
    from rdkit.Chem import GetPeriodicTable

    RDLogger.DisableLog("rdApp.*")

    periodic_table = GetPeriodicTable()
    if keep_het and data_frame_het is not None:
        data_frame = pd.concat([data_frame, data_frame_het])
    if remove_Hs:
        data_frame = data_frame[data_frame["atom_name"].str.startswith("H") ==
                                False]
        data_frame.reset_index(inplace=True, drop=True)
    x = (data_frame["x_coord"].to_numpy()
         if "x_coord" in data_frame else data_frame["x"].to_numpy())
    y = (data_frame["y_coord"].to_numpy()
         if "y_coord" in data_frame else data_frame["y"].to_numpy())
    z = (data_frame["z_coord"].to_numpy()
         if "z_coord" in data_frame else data_frame["z"].to_numpy())
    x = np.expand_dims(x, axis=1)
    y = np.expand_dims(y, axis=1)
    z = np.expand_dims(z, axis=1)
    coord = np.concatenate((x, y, z), axis=1)
    atom_type = data_frame["atom_name"].to_numpy()
    ret_atom_type, ret_coord = [], []
    for i, atom in enumerate(atom_type):
        try:
            ret_atom_type.append(periodic_table.GetAtomicNumber(atom[0]))
            ret_coord.append(coord[i])
        except:
            continue
    return np.array(ret_coord), np.array(ret_atom_type)


def general_load(name, path, sep):
    """a wrapper to download, process and load any pandas dataframe files

    Args:
        name (str): the dataset name
        path (str): the data save path

    Returns:
        pandas.DataFrame: data frame
    """
    name = download_wrapper(name, path, name)
    print_sys("Loading...")
    df = pd.read_csv(os.path.join(path, name + "." + name2type[name]), sep=sep)
    return df



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import numpy as np
from sklearn.model_selection import train_test_split
import sys

from ..utils import print_sys
from .single_cell import CellXGeneTemplate
from ..dataset_configs.config_map import scperturb_gene_datasets


def parse_single_pert(i):
    a = i.split('+')[0]
    b = i.split('+')[1]
    if a == 'ctrl':
        pert = b
    else:
        pert = a
    return pert


def parse_combo_pert(i):
    return i.split('+')[0], i.split('+')[1]


def parse_any_pert(p):
    if ('ctrl' in p) and (p != 'ctrl'):
        return [parse_single_pert(p)]
    elif 'ctrl' not in p:
        out = parse_combo_pert(p)
        return [out[0], out[1]]


def rank_genes_groups_by_cov(
    adata,
    groupby,
    control_group,
    covariate,
    pool_doses=False,
    n_genes=50,
    rankby_abs=True,
    key_added='rank_genes_groups_cov',
    return_dict=False,
):
    import scanpy as sc
    import pandas as pd
    gene_dict = {}
    cov_categories = adata.obs[covariate].unique()
    for cov_cat in cov_categories:
        #name of the control group in the groupby obs column
        control_group_cov = '_'.join([cov_cat, control_group])

        #subset adata to cells belonging to a covariate category
        adata_cov = adata[adata.obs[covariate] == cov_cat]

        #compute DEGs
        sc.tl.rank_genes_groups(adata_cov,
                                groupby=groupby,
                                reference=control_group_cov,
                                rankby_abs=rankby_abs,
                                n_genes=n_genes,
                                use_raw=False)

        #add entries to dictionary of gene sets
        de_genes = pd.DataFrame(adata_cov.uns['rank_genes_groups']['names'])
        for group in de_genes:
            gene_dict[group] = de_genes[group].tolist()

    adata.uns[key_added] = gene_dict

    if return_dict:
        return gene_dict


def get_DE_genes(adata):
    import scanpy as sc
    adata.obs.loc[:, 'dose_val'] = adata.obs.condition.apply(
        lambda x: '1+1' if len(x.split('+')) == 2 else '1')
    adata.obs.loc[:, 'control'] = adata.obs.condition.apply(
        lambda x: 0 if len(x.split('+')) == 2 else 1)
    adata.obs.loc[:, 'condition_name'] = adata.obs.apply(
        lambda x: '_'.join([x.cell_type, x.condition, x.dose_val]), axis=1)

    adata.obs = adata.obs.astype('category')
    rank_genes_groups_by_cov(adata,
                             groupby='condition_name',
                             covariate='cell_type',
                             control_group='ctrl_1',
                             n_genes=len(adata.var),
                             key_added='rank_genes_groups_cov_all')
    return adata


class PerturbOutcome(CellXGeneTemplate):

    def __init__(self, name, path="./data", print_stats=False):
        super().__init__(name, path, print_stats)
        if name in scperturb_gene_datasets:
            self.is_gene = True
        else:
            self.is_gene = False

        if name == 'scperturb_gene_NormanWeissman2019':
            self.is_combo = True
        else:
            self.is_combo = False

        if self.is_gene:
            import scanpy as sc
            sc.pp.normalize_total(self.adata)
            sc.pp.log1p(self.adata)
            from scipy.sparse import csr_matrix
            self.adata.X = csr_matrix(self.adata.X)
            sc.pp.highly_variable_genes(self.adata,
                                        n_top_genes=5000,
                                        subset=True)
            if self.is_combo:

                def map_name(x):
                    if x == 'control':
                        return 'ctrl'
                    else:
                        return '+'.join(
                            x.split('_')) if '_' in x else x + '+ctrl'

                self.adata.obs['condition'] = self.adata.obs.perturbation.apply(
                    lambda x: map_name(x))

            else:
                self.adata.obs['condition'] = self.adata.obs.perturbation.apply(
                    lambda x: x + '+ctrl' if x != 'control' else 'ctrl')
            self.adata.obs['cell_type'] = self.adata.obs['cell_line']
            self.adata.var['gene_name'] = self.adata.var.index.values
            self.adata = get_DE_genes(self.adata)

            sc.pp.highly_variable_genes(self.adata,
                                        n_top_genes=5000,
                                        subset=True)

    def get_mean_expression(self):
        raise ValueError("TODO")

    def get_DE_genes(self):
        raise ValueError("TODO")

    def get_dropout_genes(self):
        raise ValueError("TODO")

    def get_cellline_split(self,
                           ratios=[0.8, 0.1, 0.1],
                           random_state=42,
                           split_to_unseen=False,
                           remove_unseen=True):
        df = self.get_data()
        cell_line_groups = df.groupby("cell_line")
        cell_line_splits = {}
        for cell_line, cell_line_group in cell_line_groups:
            control = cell_line_group[cell_line_group["perturbation"] ==
                                      "control"]
            cell_line_group = cell_line_group[cell_line_group["perturbation"] !=
                                              "control"]
            if not split_to_unseen:
                train, tmp = train_test_split(cell_line_group,
                                              test_size=ratios[1] + ratios[2],
                                              random_state=random_state)
                test, dev = train_test_split(tmp,
                                             test_size=ratios[2] /
                                             (ratios[1] + ratios[2]),
                                             random_state=random_state)
                filter_test = test["perturbation"].isin(train["perturbation"])
                filter_dev = dev["perturbation"].isin(train["perturbation"])
                adj = 0
                if remove_unseen:
                    lbef = len(test), len(dev)
                    test = test[~filter_test]
                    dev = dev[~filter_dev]
                    laft = len(test), len(dev)
                    adj = sum(lbef) - sum(laft)
                # TODO: filters might dilute test/dev siginificantly ...
                cell_line_splits[cell_line] = {
                    "control": control,
                    "train": train,
                    "test": test,
                    "dev": dev,
                    "adj": adj,
                }
            else:
                perturbs = cell_line_group["perturbation"].unique()
                perturbs_train, tmp = train_test_split(
                    perturbs,
                    test_size=ratios[1] + ratios[2],
                    random_state=random_state)
                perturbs_test, perturbs_dev = train_test_split(
                    tmp,
                    test_size=ratios[2] / (ratios[1] + ratios[2]),
                    random_state=random_state)
                cell_line_splits[cell_line] = {
                    "control":
                        control,
                    "train":
                        cell_line_group[
                            cell_line_group["perturbation"].isin(perturbs_train)
                        ],
                    "test":
                        cell_line_group[
                            cell_line_group["perturbation"].isin(perturbs_test)
                        ],
                    "dev":
                        cell_line_group[
                            cell_line_group["perturbation"].isin(perturbs_dev)]
                }

        return cell_line_splits

    def get_perts_from_genes(self, genes, pert_list, type_='both'):
        """
            Returns all single/combo/both perturbations that include a gene
            """

        single_perts = [p for p in pert_list if ('ctrl' in p) and (p != 'ctrl')]
        combo_perts = [p for p in pert_list if 'ctrl' not in p]

        perts = []

        if type_ == 'single':
            pert_candidate_list = single_perts
        elif type_ == 'combo':
            pert_candidate_list = combo_perts
        elif type_ == 'both':
            pert_candidate_list = pert_list

        for p in pert_candidate_list:
            for g in genes:
                if g in parse_any_pert(p):
                    perts.append(p)
                    break
        return perts

    def get_genes_from_perts(self, perts):
        """
        Returns list of genes involved in a given perturbation list
        """

        if type(perts) is str:
            perts = [perts]
        gene_list = [p.split('+') for p in np.unique(perts)]
        gene_list = [item for sublist in gene_list for item in sublist]
        gene_list = [g for g in gene_list if g != 'ctrl']
        return np.unique(gene_list)

    def get_simulation_split_single(self,
                                    pert_list,
                                    train_gene_set_size=0.85,
                                    seed=1):
        unique_pert_genes = self.get_genes_from_perts(pert_list)

        pert_train = []
        pert_test = []
        np.random.seed(seed=seed)

        ## a pre-specified list of genes
        train_gene_candidates = np.random.choice(
            unique_pert_genes,
            int(len(unique_pert_genes) * train_gene_set_size),
            replace=False)

        ## ood genes
        ood_genes = np.setdiff1d(unique_pert_genes, train_gene_candidates)

        pert_single_train = self.get_perts_from_genes(train_gene_candidates,
                                                      pert_list, 'single')
        unseen_single = self.get_perts_from_genes(ood_genes, pert_list,
                                                  'single')

        return pert_single_train, unseen_single, {
            'unseen_single': unseen_single
        }

    def get_simulation_split(self,
                             pert_list,
                             train_gene_set_size=0.85,
                             combo_seen2_train_frac=0.85,
                             seed=1):

        unique_pert_genes = self.get_genes_from_perts(pert_list)

        pert_train = []
        pert_test = []
        np.random.seed(seed=seed)

        ## a pre-specified list of genes
        train_gene_candidates = np.random.choice(
            unique_pert_genes,
            int(len(unique_pert_genes) * train_gene_set_size),
            replace=False)

        ## ood genes
        ood_genes = np.setdiff1d(unique_pert_genes, train_gene_candidates)

        pert_single_train = self.get_perts_from_genes(train_gene_candidates,
                                                      pert_list, 'single')
        pert_combo = self.get_perts_from_genes(train_gene_candidates, pert_list,
                                               'combo')
        pert_train.extend(pert_single_train)

        ## the combo set with one of them in OOD
        combo_seen1 = [
            x for x in pert_combo
            if len([t for t in x.split('+') if t in train_gene_candidates]) == 1
        ]
        pert_test.extend(combo_seen1)

        pert_combo = np.setdiff1d(pert_combo, combo_seen1)
        ## randomly sample the combo seen 2 as a test set, the rest in training set
        np.random.seed(seed=seed)
        pert_combo_train = np.random.choice(
            pert_combo,
            int(len(pert_combo) * combo_seen2_train_frac),
            replace=False)

        combo_seen2 = np.setdiff1d(pert_combo, pert_combo_train).tolist()
        pert_test.extend(combo_seen2)
        pert_train.extend(pert_combo_train)

        ## unseen single
        unseen_single = self.get_perts_from_genes(ood_genes, pert_list,
                                                  'single')
        combo_ood = self.get_perts_from_genes(ood_genes, pert_list, 'combo')
        pert_test.extend(unseen_single)

        ## here only keeps the seen 0, since seen 1 is tackled above
        combo_seen0 = [
            x for x in combo_ood
            if len([t for t in x.split('+') if t in train_gene_candidates]) == 0
        ]
        pert_test.extend(combo_seen0)
        #assert len(combo_seen1) + len(combo_seen0) + len(unseen_single) + len(pert_train) + len(combo_seen2) == len(pert_list)

        return pert_train, pert_test, {
            'combo_seen0': combo_seen0,
            'combo_seen1': combo_seen1,
            'combo_seen2': combo_seen2,
            'unseen_single': unseen_single
        }

    def get_split(self,
                  ratios=[0.8, 0.1, 0.1],
                  unseen=False,
                  use_random=False,
                  random_state=42,
                  train_val_gene_set_size=0.75,
                  combo_seen2_train_frac=0.75,
                  remove_unseen=True):
        """obtain train/dev/test splits for each cell_line
        counterfactual prediction model is trained on a single cell line and then evaluated on same cell line
        and against new cell lines
        """

        if self.is_gene:
            ## use gene perturbation data split
            # check if this data has single or combo perturbations
            train_gene_set_size = train_val_gene_set_size
            combo_seen2_train_frac = combo_seen2_train_frac

            if self.is_combo:
                unique_perts = self.adata.obs.condition.unique()
                train, test, test_subgroup = self.get_simulation_split(
                    unique_perts, train_gene_set_size, combo_seen2_train_frac,
                    random_state)
                train, val, val_subgroup = self.get_simulation_split(
                    train, 0.9, 0.9, random_state)
            else:
                unique_perts = self.adata.obs.condition.unique()

                train, test, test_subgroup = self.get_simulation_split_single(
                    unique_perts, train_gene_set_size, random_state)
                train, val, val_subgroup = self.get_simulation_split_single(
                    train, 0.9, random_state)

            map_dict = {x: 'train' for x in train}
            map_dict.update({x: 'val' for x in val})
            map_dict.update({x: 'test' for x in test})
            map_dict.update({'ctrl': 'train'})

            self.adata.obs['split'] = self.adata.obs['condition'].map(map_dict)
            adata_out = {
                "train": self.adata[self.adata.obs.split == 'train'],
                "dev": self.adata[self.adata.obs.split == 'val'],
                "test": self.adata[self.adata.obs.split == 'test']
            }
            subgroup = {
                'test_subgroup': test_subgroup,
                'dev_subgroup': val_subgroup
            }
            return adata_out, subgroup

        if not use_random:
            return self.get_cellline_split(split_to_unseen=unseen,
                                           ratios=ratios,
                                           random_state=random_state,
                                           remove_unseen=remove_unseen)
        df = self.get_data()
        # just do a random split, otherwise you'll split by cell line...
        control = df[df["perturbation"] == "control"]
        perturbs = df[df["perturbation"] != "control"]
        train, tmp = train_test_split(perturbs,
                                      test_size=ratios[1] + ratios[2],
                                      random_state=random_state)
        test, dev = train_test_split(tmp,
                                     test_size=ratios[2] /
                                     (ratios[1] + ratios[2]),
                                     random_state=random_state)
        return {"control": control, "train": train, "dev": dev, "test": test}
